from __future__ import absolute_import, print_function

import hashlib
import inspect
import itertools
import logging
import os
import sys
from abc import abstractmethod

import jinja2
import jinja2.meta
import sqlparse
from decorator import decorator
from interface_meta import quirk_docs, override

from omniduct.caches.base import cached_method
from omniduct.duct import Duct
from omniduct.filesystems.local import LocalFsClient
from omniduct.utils.debug import logger, logging_scope
from omniduct.utils.decorators import require_connection
from omniduct.utils.magics import (MagicsProvider, process_line_arguments,
                                   process_line_cell_arguments)

from . import _cursor_formatters
from ._cursor_serializer import CursorSerializer
from ._namespaces import ParsedNamespaces

logging.getLogger('requests').setLevel(logging.WARNING)


@decorator
def render_statement(method, self, statement, *args, **kwargs):
    """
    Pre-render a statement template prior to wrapped function execution.

    This decorator expects to act as wrapper on functions which
    takes statements as the second argument.
    """
    if kwargs.pop('template', True):
        statement = self.template_render(
            statement,
            context=kwargs.pop('context', {}),
            by_name=False,
        )
    return method(self, statement, *args, **kwargs)


class DatabaseClient(Duct, MagicsProvider):
    """
    An abstract class providing the common API for all database clients.

    Note: `DatabaseClient` subclasses are callable, so that one can use
    `DatabaseClient(...)` as a short-hand for `DatabaseClient.query(...)`.

    Class Attributes:
        DUCT_TYPE (`Duct.Type`): The type of `Duct` protocol implemented by this class.
        DEFAULT_PORT (int): The default port for the database service (defined
            by subclasses).
        CURSOR_FORMATTERS (dict<str, CursorFormatter): asdsd
        DEFAULT_CURSOR_FORMATTER (str): ...
    """

    DUCT_TYPE = Duct.Type.DATABASE
    DEFAULT_PORT = None

    CURSOR_FORMATTERS = {
        'pandas': _cursor_formatters.PandasCursorFormatter,
        'hive': _cursor_formatters.HiveCursorFormatter,
        'csv': _cursor_formatters.CsvCursorFormatter,
        'tuple': _cursor_formatters.TupleCursorFormatter,
        'dict': _cursor_formatters.DictCursorFormatter,
        'raw': _cursor_formatters.RawCursorFormatter,
    }
    DEFAULT_CURSOR_FORMATTER = 'pandas'
    SUPPORTS_SESSION_PROPERTIES = False
    NAMESPACE_NAMES = ['database', 'table']
    NAMESPACE_QUOTECHAR = '"'
    NAMESPACE_SEPARATOR = '.'

    @property
    def NAMESPACE_DEFAULT(self):
        pass

    @quirk_docs('_init', mro=True)
    def __init__(
        self, session_properties=None, templates=None, template_context=None, default_format_opts=None,
        **kwargs
    ):
        """
        session_properties (dict): A mapping of default session properties
            to values. Interpretation is left up to implementations.
        templates (dict): A dictionary of name to template mappings. Additional
            templates can be added using `.template_add`.
        template_context (dict): The default template context to use when
            rendering templates.
        default_format_opts (dict): The default formatting options passed to
            cursor formatter.
        """
        Duct.__init_with_kwargs__(self, kwargs, port=self.DEFAULT_PORT)

        self.session_properties = session_properties or {}
        self._templates = templates or {}
        self._template_context = template_context or {}
        self._sqlalchemy_engine = None
        self._sqlalchemy_metadata = None
        self._default_format_opts = default_format_opts or {}

        self._init(**kwargs)

    @abstractmethod
    def _init(self):
        pass

    # Session property management and configuration
    @property
    def session_properties(self):
        """dict: The default session properties used in statement executions."""
        return self._session_properties

    @session_properties.setter
    def session_properties(self, properties):
        self._session_properties = self._get_session_properties(default=properties)

    def _get_session_properties(self, overrides=None, default=None):
        """
        Retrieve the default session properties with optional overrides.

        Properties with a value of None will be skipped, in order to allow
        overrides to remove default properties.

        Args:
            overrides (dict, None): A dictionary of session property overrides.
            default (dict, None): A dictionary of default session properties, if
                it is necessary to override `self.session_properties`.

        Returns:
            dict: A dictionary of session properties.
        """
        if (default or overrides) and not self.SUPPORTS_SESSION_PROPERTIES:
            raise RuntimeError("Session properties are not supported by this backend.")

        props = (default if default is not None else self.session_properties).copy()
        props.update(overrides or {})

        # Remove any properties with value set to None.
        for key, value in props.items():
            if value is None:
                del props[key]

        return props

    def __call__(self, query, **kwargs):
        return self.query(query, **kwargs)

    # Querying
    def _statement_prepare(self, statement, session_properties, **kwargs):
        """
        Prepare a statement for execution.

        This is a hook that can be used by subclasses to transform and/or
        validate statements before execution in `_execute`; for example
        inserting session properties in query headers.

        Args:
            statement (str): The statement to be executed.
            session_properties (dict): A mutable dictionary of session properties
                and their values (this method can mutate it depending on statement
                contents).
            **kwargs (dict): Any additional keyword arguments passed through to
                `self.execute` (will match the extra keyword arguments added to
                `self._execute`).

        Returns:
            statement (str): The statement to be executed (potentially transformed).
        """
        return statement

    def _statement_split(self, statements):
        """
        Split a statement into separate SQL statements.

        This method converts a single string containing one or more SQL
        statements into an iterator of strings, each corresponding to one SQL
        statement. If the statement's language is not to be SQL, this method
        should be overloaded appropriately.

        Args:
            statements (str): A string containing one or more SQL statements.

        Returns:
            iterator<str>: An iterator of SQL statements.
        """
        for statement in sqlparse.split(statements):
            statement = statement.strip()
            if statement.endswith(';'):
                statement = statement[:-1].strip()
            if statement:  # remove empty statements
                yield statement

    @classmethod
    def statement_hash(cls, statement, cleanup=True):
        """
        Retrieve the hash to use to identify query statements to the cache.

        Args:
            statement (str): A string representation of the statement to be
                hashed.
            cleanup (bool): Whether the statement should first be consistently
                reformatted using `statement_cleanup`.

        Returns:
            str: The hash used to identify a statement to the cache.
        """
        if cleanup:
            statement = cls.statement_cleanup(statement)
        if (
            sys.version_info.major == 3
            or sys.version_info.major == 2 and isinstance(statement, unicode)  # noqa: F821
        ):
            statement = statement.encode('utf8')
        return hashlib.sha256(statement).hexdigest()

    @classmethod
    def statement_cleanup(cls, statement):
        """
        Clean up statements prior to hash computation.

        This classmethod takes an SQL statement and reformats it by consistently
        removing comments and replacing all whitespace. It is used by the
        `statement_hash` method to avoid functionally identical queries hitting
        different cache keys. If the statement's language is not to be SQL, this
        method should be overloaded appropriately.

        Args:
            statement (str): The statement to be reformatted/cleaned-up.

        Returns:
            str: The new statement, consistently reformatted.
        """
        statement = sqlparse.format(statement, strip_comments=True, reindent=True)
        statement = os.linesep.join([line for line in statement.splitlines() if line])
        return statement

    @render_statement
    @cached_method(
        key=lambda self, kwargs: self.statement_hash(
            statement=kwargs['statement'],
            cleanup=kwargs.pop('cleanup', True)
        ),
        serializer=lambda self, kwargs: CursorSerializer(),
        use_cache=lambda self, kwargs: kwargs.pop('use_cache', False),
        metadata=lambda self, kwargs: {
            'statement': kwargs['statement'],
            'session_properties': kwargs['session_properties']
        }
    )
    @quirk_docs('_execute')
    @require_connection
    def execute(self, statement, wait=True, cursor=None, session_properties=None, **kwargs):
        """
        Execute a statement against this database and return a cursor object.

        Where supported by database implementations, this cursor can the be used
        in future executions, by passing it as the `cursor` keyword argument.

        Args:
            statement (str): The statement to be executed by the query client
                (possibly templated).
            wait (bool): Whether the cursor should be returned before the
                server-side query computation is complete and the relevant
                results downloaded.
            cursor (DBAPI2 cursor):  Rather than creating a new cursor, execute
                the statement against the provided cursor.
            session_properties (dict): Additional session properties and/or
                overrides to use for this query. Setting a session property
                value to `None` will cause it to be omitted.
            **kwargs (dict): Extra keyword arguments to be passed on to
                `_execute`, as implemented by subclasses.
            template (bool): Whether the statement should be treated as a Jinja2
                template. [Used by `render_statement` decorator.]
            context (dict): The context in which the template should be
                evaluated (a dictionary of parameters to values). [Used by
                `render_statement` decorator.]
            use_cache (bool): True or False (default). Whether to use the cache
                (if present). [Used by `cached_method` decorator.]
            renew (bool): True or False (default). If cache is being used, renew
                it before returning stored value. [Used by `cached_method`
                decorator.]
            cleanup (bool): Whether statement should be cleaned up before
                computing the hash used to cache results. [Used by `cached_method`
                decorator.]

        Returns:
            DBAPI2 cursor: A DBAPI2 compatible cursor instance.
        """

        session_properties = self._get_session_properties(overrides=session_properties)

        statements = list(self._statement_split(
            self._statement_prepare(statement, session_properties=session_properties, **kwargs)
        ))
        assert len(statements) > 0, "No non-empty statements were provided."

        for statement in statements[:-1]:
            cursor = self._execute(statement, cursor=cursor, wait=True, session_properties=session_properties, **kwargs)
        cursor = self._execute(statements[-1], cursor=cursor, wait=wait, session_properties=session_properties, **kwargs)

        return cursor

    @logging_scope("Query", timed=True)
    @render_statement
    def query(self, statement, format=None, format_opts={}, use_cache=True, **kwargs):
        """
        Execute a statement against this database and collect formatted data.

        Args:
            statement (str): The statement to be executed by the query client
                (possibly templated).
            format (str): A subclass of CursorFormatter, or one of: 'pandas',
                'hive', 'csv', 'tuple' or 'dict'. Defaults to
                `self.DEFAULT_CURSOR_FORMATTER`.
            format_opts (dict): A dictionary of format-specific options.
            use_cache (bool): Whether to cache the cursor returned by
                `DatabaseClient.execute()` (overrides the default of False
                for `.execute()`). (default=True)
            **kwargs (dict): Additional arguments to pass on to
                `DatabaseClient.execute()`.

        Returns:
            The results of the query formatted as nominated.
        """
        cursor = self.execute(statement, wait=True, template=False, use_cache=use_cache, **kwargs)

        # Some DBAPI2 cursor implementations error if attempting to extract
        # data from an empty cursor, and if so, we simply return None.
        if self._cursor_empty(cursor):
            return None

        formatter = self._get_formatter(format, cursor, **format_opts)
        return formatter.dump()

    def stream(self, statement, format=None, format_opts={}, batch=None, **kwargs):
        """
        Execute a statement against this database and stream formatted results.

        This method returns a generator over objects representing rows in the
        result set. If `batch` is not `None`, then the iterator
        will be over lists of length `batch` containing formatted rows.

        Args:
            statement (str): The statement to be executed against the database.
            format (str): A subclass of CursorFormatter, or one of: 'pandas',
                'hive', 'csv', 'tuple' or 'dict'. Defaults to
                `self.DEFAULT_CURSOR_FORMATTER`.
            format_opts (dict): A dictionary of format-specific options.
            batch (int): If not `None`, the number of rows from the resulting
                cursor to be returned at once.
            **kwargs (dict): Additional keyword arguments to pass onto
                `DatabaseClient.execute`.

        Returns:
            iterator: An iterator over objects of the nominated format or, if
                batched, a list of such objects.
        """
        cursor = self.execute(statement, wait=True, **kwargs)
        formatter = self._get_formatter(format, cursor, **format_opts)

        for row in formatter.stream(batch=batch):
            yield row

    def _get_formatter(self, formatter, cursor, **kwargs):
        formatter = formatter or self.DEFAULT_CURSOR_FORMATTER
        if not (inspect.isclass(formatter) and issubclass(formatter, _cursor_formatters.CursorFormatter)):
            assert formatter in self.CURSOR_FORMATTERS, "Invalid format '{}'. Choose from: {}".format(formatter, ','.join(self.CURSOR_FORMATTERS.keys()))
            formatter = self.CURSOR_FORMATTERS[formatter]
        format_opts = dict(itertools.chain(self._default_format_opts.items(), kwargs.items()))
        return formatter(cursor, **format_opts)

    def stream_to_file(self, statement, file, format='csv', fs=None, **kwargs):
        """
        Execute a statement against this database and stream results to a file.

        This method is a wrapper around `DatabaseClient.stream` that enables the
        iterative writing of cursor results to a file. This is especially useful
        when there are a very large number of results, and loading them all into
        memory would require considerable resources. Note that 'csv' is the
        default format for this method (rather than `pandas`).

        Args:
            statement (str): The statement to be executed against the database.
            file (str, file-like-object): The filename where the data should be
                written, or an open file-like resource.
            format (str): The format to be used ('csv' by default). Format
                options can be passed via `**kwargs`.
            fs (None, FileSystemClient): The filesystem wihin which the
                nominated file should be found. If `None`, the local filesystem
                will be used.
            **kwargs: Additional keyword arguments to pass onto
                `DatabaseClient.stream`.
        """
        close_later = False
        if isinstance(file, str):
            file = (fs or LocalFsClient()).open(file, 'w')
            close_later = True

        try:
            file.writelines(self.stream(statement, format=format, **kwargs))
        finally:
            if close_later:
                file.close()

    def execute_from_file(self, file, fs=None, **kwargs):
        """
        Execute a statement stored in a file.

        Args:
            file (str, file-like-object): The path of the file containing the
                query statement to be executed against the database, or an open
                file-like resource.
            fs (None, FileSystemClient): The filesystem wihin which the
                nominated file should be found. If `None`, the local filesystem
                will be used.
            **kwargs (dict): Extra keyword arguments to pass on to
                `DatabaseClient.execute`.

        Returns:
            DBAPI2 cursor: A DBAPI2 compatible cursor instance.
        """
        close_later = False
        if isinstance(file, str):
            file = (fs or LocalFsClient()).open(file, 'r')
            close_later = True

        try:
            return self.execute(file.read(), **kwargs)
        finally:
            if close_later:
                file.close()

    def query_from_file(self, file, fs=None, **kwargs):
        """
        Query using a statement stored in a file.

        Args:
            file (str, file-like-object): The path of the file containing the
                query statement to be executed against the database, or an open
                file-like resource.
            fs (None, FileSystemClient): The filesystem wihin which the
                nominated file should be found. If `None`, the local filesystem
                will be used.
            **kwargs (dict): Extra keyword arguments to pass on to
                `DatabaseClient.query`.

        Returns:
            object: The results of the query formatted as nominated.
        """
        close_later = False
        if isinstance(file, str):
            file = (fs or LocalFsClient()).open(file, 'r')
            close_later = True

        try:
            return self.query(file.read(), **kwargs)
        finally:
            if close_later:
                file.close()

    @property
    def template_names(self):
        """
        list: A list of names associated with the templates associated with this
        client.
        """
        return list(self._templates)

    def template_add(self, name, body):
        """
        Add a named template to the internal dictionary of templates.

        Note: Templates are interpreted as `jinja2` templates. See
        `.template_render` for more information.

        Args:
            name (str): The name of the template.
            body (str): The (typically) multiline body of the template.

        Returns:
            PrestoClient: A reference to this object.
        """
        self._templates[name] = body
        return self

    def template_get(self, name):
        """
        Retrieve a named template.

        Args:
            name (str): The name of the template to retrieve.

        Raises:
            ValueError: If `name` is not associated with a template.

        Returns:
            str: The requested template.
        """
        if name not in self._templates:
            raise ValueError("No such template named: '{}'.".format(name))
        return self._templates[name]

    def template_variables(self, name_or_statement, by_name=False):
        """
        Return the set of undeclared variables required for this template.

        Args:
            name_or_statement (str): The name of a template (if `by_name` is True)
                or else a string representation of a `jinja2` template.
            by_name (bool): `True` if `name_or_statement` should be interpreted as a
                template name, or `False` (default) if `name_or_statement` should be
                interpreted as a template body.

        Returns:
            set<str>: A set of names which the template requires to be rendered.
        """
        ast = jinja2.Environment().parse(
            self.template_render(name_or_statement, by_name=by_name, meta_only=True)
        )
        return jinja2.meta.find_undeclared_variables(ast)

    def template_render(self, name_or_statement, context=None, by_name=False,
                        cleanup=False, meta_only=False):
        """
        Render a template by name or value.

        In addition to the `jinja2` templating syntax, described in more detail
        in the official `jinja2` documentation, a meta-templating extension is
        also provided. This meta-templating allows you to reference other
        reference other templates. For example, if you had two SQL templates
        named 'template_a' and 'template_b', then you could render them into one
        SQL query using (for example):

        .. code-block:: python

            .template_render('''
            WITH
                a AS (
                    {{{template_a}}}
                ),
                b AS (
                    {{{template_b}}}
                )
            SELECT *
            FROM a
            JOIN b ON a.x = b.x
            ''')

        Note that template substitution in this way is iterative, so you can
        chain template embedding, provided that such embedding is not recursive.

        Args:
            name_or_statement (str): The name of a template (if `by_name` is True)
                or else a string representation of a `jinja2` template.
            context (dict, None): A dictionary to use as the template context.
                If not specified, an empty dictionary is used.
            by_name (bool): `True` if `name_or_statement` should be interpreted as a
                template name, or `False` (default) if `name_or_statement` should be
                interpreted as a template body.
            cleanup (bool): `True` if the rendered statement should be formatted,
                `False` (default) otherwise
            meta_only (bool): `True` if rendering should only progress as far as
                rendering nested templates (i.e. don't actually substitute in
                variables from the context); `False` (default) otherwise.

        Returns:
            str: The rendered template.
        """
        if by_name:
            if name_or_statement not in self._templates:
                raise ValueError("No such template of name: '{}'.".format(name_or_statement))
            statement = self._templates[name_or_statement]
        else:
            statement = name_or_statement

        try:
            from sqlalchemy.sql.base import Executable
            if isinstance(statement, Executable):
                statement = str(statement.compile(compile_kwargs={"literal_binds": True}))
        except ImportError:
            pass

        if context is None or context is False:
            context = {}

        template_context = {}
        template_context.update(self._template_context)  # default context
        template_context.update(context)  # context passed in
        intersection = set(self._template_context.keys()) & set(context.keys())
        if intersection:
            logger.warning(
                "The following default template context keys have been overridden "
                "by the local context: {}."
                .format(intersection)
            )

        # Substitute in any other named statements recursively
        while '{{{' in statement or '{{%' in statement:
            statement = (
                jinja2.Template(
                    statement,
                    block_start_string='{{%',
                    block_end_string='%}}',
                    variable_start_string='{{{',
                    variable_end_string='}}}',
                    comment_start_string='{{#',
                    comment_end_string='#}}',
                    undefined=jinja2.StrictUndefined
                )
                .render(getattr(self, '_templates', {}))
            )

        if not meta_only:
            statement = (
                jinja2.Template(statement, undefined=jinja2.StrictUndefined)
                .render(template_context)
            )

        if cleanup:
            statement = self.statement_cleanup(statement)

        return statement

    def execute_from_template(self, name, context=None, **kwargs):
        """
        Render and then execute a named template.

        Args:
            name (str): The name of the template to be rendered and executed.
            context (dict): The context in which the template should be rendered.
            **kwargs (dict): Additional parameters to pass to `.execute()`.

        Returns:
            DBAPI2 cursor: A DBAPI2 compatible cursor instance.
        """
        statement = self.template_render(name, context, by_name=True)
        return self.execute(statement, **kwargs)

    def query_from_template(self, name, context=None, **kwargs):
        """
        Render and then query using a named tempalte.

        Args:
            name (str): The name of the template to be rendered and used to query
                the database.
            context (dict): The context in which the template should be rendered.
            **kwargs (dict): Additional parameters to pass to `.query()`.

        Returns:
           object: The results of the query formatted as nominated.
        """
        statement = self.template_render(name, context, by_name=True)
        return self.query(statement, **kwargs)

    # Uploading/querying data into data store
    @logging_scope('Query [CTAS]', timed=True)
    @quirk_docs('_query_to_table')
    def query_to_table(self, statement, table, if_exists='fail', **kwargs):
        """
        Run a query and store the results in a table in this database.

        Args:
            statement: The statement to be executed.
            table (str): The name of the table into which the dataframe should
                be uploaded.
            if_exists (str): if nominated table already exists: 'fail' to do
                nothing, 'replace' to drop, recreate and insert data into new
                table, and 'append' to add data from this table into the
                existing table.
            **kwargs (dict): Additional keyword arguments to pass onto
                `DatabaseClient._query_to_table`.

        Returns:
            DB-API cursor: The cursor object associated with the execution.
        """
        assert if_exists in {'fail', 'replace', 'append'}
        table = self._parse_namespaces(table)
        return self._query_to_table(statement, table, if_exists=if_exists, **kwargs)

    @logging_scope('Dataframe Upload', timed=True)
    @quirk_docs('_dataframe_to_table')
    @require_connection
    def dataframe_to_table(self, df, table, if_exists='fail', **kwargs):
        """
        Upload a local pandas dataframe into a table in this database.

        Args:
            df (pandas.DataFrame): The dataframe to upload into the database.
            table (str, ParsedNamespaces): The name of the table into which the
                dataframe should be uploaded.
            if_exists (str): if nominated table already exists: 'fail' to do
                nothing, 'replace' to drop, recreate and insert data into new
                table, and 'append' to add data from this table into the
                existing table.
            **kwargs (dict): Additional keyword arguments to pass onto
                `DatabaseClient._dataframe_to_table`.
        """
        assert if_exists in {'fail', 'replace', 'append'}
        self._dataframe_to_table(df, self._parse_namespaces(table), if_exists=if_exists, **kwargs)

    # Table properties

    @abstractmethod
    def _execute(self, statement, cursor, wait, session_properties):
        pass

    def _query_to_table(self, statement, table, if_exists, **kwargs):
        raise NotImplementedError

    def _dataframe_to_table(self, df, table, if_exists='fail', **kwargs):
        raise NotImplementedError

    def _cursor_empty(self, cursor):
        return cursor is None

    def _parse_namespaces(self, name, level=0, defaults=None):
        return ParsedNamespaces.from_name(
            name,
            self.NAMESPACE_NAMES[:-level] if level > 0 else self.NAMESPACE_NAMES,
            quote_char=self.NAMESPACE_QUOTECHAR,
            separator=self.NAMESPACE_SEPARATOR,
            defaults=self.NAMESPACE_DEFAULT if defaults is None else defaults
        )

    @quirk_docs('_table_list')
    def table_list(self, namespace=None, renew=True, **kwargs):
        """
        Return a list of table names in the data source as a DataFrame.

        Args:
            namespace (str): The namespace in which to look for tables.
            renew (bool): Whether to renew the table list or use cached results
                (default: True).
            **kwargs (dict): Additional arguments passed through to implementation.

        Returns:
            list<str>: The names of schemas in this database.
        """
        return self._table_list(self._parse_namespaces(namespace, level=1), renew=renew, **kwargs)

    @abstractmethod
    def _table_list(self, namespace, **kwargs):
        pass

    @quirk_docs('_table_exists')
    def table_exists(self, table, renew=True, **kwargs):
        """
        Check whether a table exists.

        Args:
            table (str): The table for which to check.
            renew (bool): Whether to renew the table list or use cached results
                (default: True).
            **kwargs (dict): Additional arguments passed through to implementation.

        Returns:
            bool: `True` if table exists, and `False` otherwise.
        """
        return self._table_exists(table=self._parse_namespaces(table), renew=renew, **kwargs)

    @abstractmethod
    def _table_exists(self, table, **kwargs):
        pass

    @quirk_docs('_table_drop')
    def table_drop(self, table, **kwargs):
        """
        Remove a table from the database.

        Args:
            table (str): The table to drop.
            **kwargs (dict): Additional arguments passed through to implementation.

        Returns:
            DB-API cursor: The cursor associated with this execution.
        """
        return self._table_drop(table=self._parse_namespaces(table), **kwargs)

    @abstractmethod
    def _table_drop(self, table, **kwargs):
        pass

    @quirk_docs('_table_desc')
    def table_desc(self, table, renew=True, **kwargs):
        """
        Describe a table in the database.

        Args:
            table (str): The table to describe.
            **kwargs (dict): Additional arguments passed through to implementation.

        Returns:
            pandas.DataFrame: A dataframe description of the table.
        """
        return self._table_desc(table=self._parse_namespaces(table), renew=renew, **kwargs)

    @abstractmethod
    def _table_desc(self, table, **kwargs):
        pass

    @quirk_docs('_table_head')
    def table_head(self, table, n=10, renew=True, **kwargs):
        """
        Retrieve the first `n` rows from a table.

        Args:
            table (str): The table from which to extract data.
            n (int): The number of rows to extract.
            renew (bool): Whether to renew the table list or use cached results
                (default: True).
            **kwargs (dict): Additional arguments passed through to implementation.

        Returns:
            pandas.DataFrame: A dataframe representation of the first `n` rows
                of the nominated table.
        """
        return self._table_head(table=self._parse_namespaces(table), n=n, renew=renew, **kwargs)

    @abstractmethod
    def _table_head(self, table, n=10, **kwargs):
        pass

    @quirk_docs('_table_props')
    def table_props(self, table, renew=True, **kwargs):
        """
        Retrieve the properties associated with a table.

        Args:
            table (str): The table from which to extract data.
            renew (bool): Whether to renew the table list or use cached results
                (default: True).
            **kwargs (dict): Additional arguments passed through to implementation.

        Returns:
            pandas.DataFrame: A dataframe representation of the table
                properties.
        """
        return self._table_props(table=self._parse_namespaces(table), renew=renew, **kwargs)

    @abstractmethod
    def _table_props(self, table, **kwargs):
        pass

    @override
    def _register_magics(self, base_name):
        """
        The following magic functions will be registered (assuming that
        the base name is chosen to be 'hive'):
        - Cell Magics:
            - `%%hive`: For querying the database.
            - `%%hive.execute`: For executing a statement against the database.
            - `%%hive.stream`: For executing a statement against the database,
                and streaming the results.
            - `%%hive.template`: The defining a new template.
            - `%%hive.render`: Render a provided query statement.
        - Line Magics:
            - `%hive`: For querying the database using a named template.
            - `%hive.execute`: For executing a named template statement against
                the database.
            - `%hive.stream`: For executing a named template against the database,
                and streaming the results.
            - `%hive.render`: Render a provided a named template.
            - `%hive.desc`: Describe the table nominated.
            - `%hive.head`: Return the first rows in a specified table.
            - `%hive.props`: Show the properties specified for a nominated table.

        Documentation for these magics is provided online.
        """
        from IPython import get_ipython
        from IPython.core.magic import register_line_magic, register_cell_magic, register_line_cell_magic

        def statement_executor_magic(executor, statement, variable=None, show='head', transpose=False, template=True, context=None, **kwargs):

            ip = get_ipython()

            if context is None:
                context = ip.user_ns

            # Line magic
            if statement is None:
                return self.query_from_template(variable, context=context, **kwargs)

            # Cell magic
            result = getattr(self, executor)(statement, template=template, context=context, **kwargs)

            if variable is not None:
                ip.user_ns[variable] = result

            if executor != 'query':
                if variable is None:
                    return result
                return
            elif variable is None:
                return result

            format = kwargs.get('format', self.DEFAULT_CURSOR_FORMATTER)
            if show == 'head':
                show = 10
            if isinstance(show, int):
                r = result.head(show) if format == 'pandas' else result[:show]
            elif show == 'all':
                r = result
            elif show in (None, 'none'):
                return None
            else:
                raise ValueError("Omniduct does not recognise the argument show='{0}' in cell magic.".format(show))

            if format == 'pandas' and transpose:
                return r.T
            return r

        @register_line_cell_magic(base_name)
        @process_line_cell_arguments
        def query_magic(*args, **kwargs):
            return statement_executor_magic('query', *args, **kwargs)

        @register_line_cell_magic("{}.{}".format(base_name, 'execute'))
        @process_line_cell_arguments
        def execute_magic(*args, **kwargs):
            return statement_executor_magic('execute', *args, **kwargs)

        @register_line_cell_magic("{}.{}".format(base_name, 'stream'))
        @process_line_cell_arguments
        def stream_magic(*args, **kwargs):
            return statement_executor_magic('stream', *args, **kwargs)

        @register_cell_magic("{}.{}".format(base_name, 'template'))
        @process_line_arguments
        def template_add(body, name):
            self.template_add(name, body)

        @register_line_cell_magic("{}.{}".format(base_name, 'render'))
        @process_line_cell_arguments
        def template_render_magic(body=None, name=None, context=None, show=True,
                                  cleanup=False, meta_only=False):

            ip = get_ipython()

            if body is None:
                assert name is not None, "Name must be specified in line-mode."
                rendered = self.template_render(
                    name, context=context or ip.user_ns, by_name=True,
                    cleanup=cleanup, meta_only=meta_only
                )
            else:
                rendered = self.template_render(
                    body, context=context or ip.user_ns, by_name=False,
                    cleanup=cleanup, meta_only=meta_only
                )
                if name is not None:
                    ip.user_ns[name] = rendered

            if show:
                print(rendered)
            else:
                return rendered

        @register_line_magic("{}.{}".format(base_name, 'desc'))
        @process_line_arguments
        def table_desc(table_name, **kwargs):
            return self.table_desc(table_name, **kwargs)

        @register_line_magic("{}.{}".format(base_name, 'head'))
        @process_line_arguments
        def table_head(table_name, **kwargs):
            return self.table_head(table_name, **kwargs)

        @register_line_magic("{}.{}".format(base_name, 'props'))
        @process_line_arguments
        def table_props(table_name, **kwargs):
            return self.table_props(table_name, **kwargs)



from __future__ import absolute_import

import pandas as pd
import sqlalchemy
from sqlalchemy import Table
from sqlalchemy import types as sql_types

from omniduct.utils.debug import logger
from omniduct.utils.decorators import require_connection

try:
    from pyhive.sqlalchemy_presto import PrestoDialect

    def get_columns(self, connection, table_name, schema=None, **kw):
        # Extend types supported by PrestoDialect as defined in PyHive
        type_map = {
            'bigint': sql_types.BigInteger,
            'integer': sql_types.Integer,
            'boolean': sql_types.Boolean,
            'double': sql_types.Float,
            'varchar': sql_types.String,
            'timestamp': sql_types.TIMESTAMP,
            'date': sql_types.DATE,
            'array<bigint>': sql_types.ARRAY(sql_types.Integer),
            'array<varchar>': sql_types.ARRAY(sql_types.String)
        }

        rows = self._get_table_columns(connection, table_name, schema)
        result = []
        for row in rows:
            try:
                coltype = type_map[row.Type]
            except KeyError:
                logger.warn("Did not recognize type '%s' of column '%s'" % (row.Type, row.Column))
                coltype = sql_types.NullType
            result.append({
                'name': row.Column,
                'type': coltype,
                # newer Presto no longer includes this column
                'nullable': getattr(row, 'Null', True),
                'default': None,
            })
        return result

    PrestoDialect.get_columns = get_columns
except ImportError:
    logger.debug("Not monkey patching pyhive's PrestoDialect.get_columns due to missing dependencies.")


class SchemasMixin(object):
    """
    Attaches a tab-completable `.schemas` attribute to a `DatabaseClient` instance.

    It is currently implemented as a mixin rather than directly provided on the
    base class because it requires that the host `DatabaseClient` instance have a
    `sqlalchemy` metadata object handle, and not all backends support this.

    If we are willing to forgo the ability to actually make queries using the
    SQLAlchemy ORM, we could instead use an SQL agnostic version.
    """

    @property
    @require_connection
    def schemas(self):
        """
        object: An object with attributes corresponding to the names of the schemas
            in this database.
        """
        from werkzeug import LocalProxy

        def get_schemas():
            if not getattr(self, '_schemas', None):
                assert getattr(self, '_sqlalchemy_metadata', None) is not None, (
                    "`{class_name}` instances do not provide the required sqlalchemy metadata "
                    "for schema exploration.".format(self.__class__.__name__)
                )
                self._schemas = Schemas(self._sqlalchemy_metadata)
            return self._schemas
        return LocalProxy(get_schemas)


# Extend Table to support returning pandas description of table
class TableDesc(Table):
    """
    Extends the SQL Alchemy `Table` class with some short-hand introspection methods.
    """

    def desc(self):
        """pandas.DataFrame: The description of this SQL table."""
        return pd.DataFrame(
            [[col.name, col.type.compile(self.bind.dialect)] for col in self.columns.values()],
            columns=['name', 'type']
        )

    def head(self, n=10):
        """
        Retrieve the first `n` rows from this table.

        Args:
            n (int): The number of rows to retrieve from this table.

        Returns:
            pandas.DataFrame: A dataframe representation of the first `n` rows
                of this table.
        """
        statement = self.select()
        if n is not None:
            statement = statement.limit(n)
        return pd.read_sql(statement, self.bind)

    def dump(self):
        """
        Retrieve the entire database table as a pandas DataFrame.

        Returns:
            pandas.DataFrame: A dataframe representation of the entire table.
        """
        return self.head(n=None)

    def __repr__(self):
        return self.desc().__repr__()


# Define helpers to allow for table completion/etc
class Schemas(object):
    """
    An object which has as its attributes all of the schemas in a nominated database.

    Args:
        metadata (sqlalchemy.MetaData): A SQL Alchemy `MetaData` instance
            configured for the nominated database.
    """

    def __init__(self, metadata):
        self._metadata = metadata
        self._schema_names = None
        self._schema_cache = {}

    @property
    def all(self):
        "list<str>: The list of schema names."
        if self._schema_names is None:
            self._schema_names = sqlalchemy.inspect(self._metadata.bind).get_schema_names()
        return self._schema_names

    def __dir__(self):
        return self.all

    def __getattr__(self, value):
        if value in self.all:
            if value not in self._schema_cache:
                self._schema_cache[value] = Schema(metadata=self._metadata, schema=value)
            return self._schema_cache[value]
        raise AttributeError("No such schema {}".format(value))

    def __repr__(self):
        return "<Schemas: {} schemas>".format(len(self.all))

    def __iter__(self):
        for schema in self.all:
            yield schema

    def __len__(self):
        return len(self.all)


class Schema(object):
    """
    An object which has as its attributes all of the tables in a nominated database schema.

    Args:
        metadata (sqlalchemy.MetaData): A SQL Alchemy `MetaData` instance
            configured for the nominated database.
        schema (str): The schema within which to expose tables.
    """

    def __init__(self, metadata, schema):
        self._metadata = metadata
        self._schema = schema
        self._table_cache = {}
        self._table_names = None

    @property
    def all(self):
        """list<str>: The table names in this database schema."""
        if self._table_names is None:
            self._table_names = sqlalchemy.inspect(self._metadata.bind).get_table_names(self._schema)
        return self._table_names

    def __dir__(self):
        return self.all

    def __getattr__(self, table):
        if table in self.all:
            if table not in self._table_cache:
                self._table_cache[table] = TableDesc(
                    '{}'.format(table), self._metadata, autoload=True, schema=self._schema
                )
            return self._table_cache[table]
        raise AttributeError("No such table {}".format(table))

    def __repr__(self):
        return "<Schema `{}`: {} tables>".format(self._schema, len(self.all))

    def __iter__(self):
        for schema in self.all:
            yield schema

    def __len__(self):
        return len(self.all)



from pyarrow.filesystem import FileSystem
from pyarrow.util import implements, _stringify_path


class OmniductFileSystem(FileSystem):
    """
    Wraps Omniduct filesystem implementations for use with PyArrow.
    """

    def __init__(self, fs):
        self.fs = fs

    @implements(FileSystem.isdir)
    def isdir(self, path):
        return self.fs.isdir(_stringify_path(path))

    @implements(FileSystem.isfile)
    def isfile(self, path):
        return self.fs.isfile(_stringify_path(path))

    @implements(FileSystem._isfilestore)
    def _isfilestore(self):
        return True

    @implements(FileSystem.delete)
    def delete(self, path, recursive=False):
        return self.fs.remove(_stringify_path(path), recursive=recursive)

    @implements(FileSystem.exists)
    def exists(self, path):
        return self.fs.exists(_stringify_path(path))

    @implements(FileSystem.mkdir)
    def mkdir(self, path, create_parents=True):
        return self.fs.mkdir(_stringify_path(path), recursive=create_parents)

    @implements(FileSystem.open)
    def open(self, path, mode='rb'):
        return self.fs.open(_stringify_path(path), mode=mode)

    @implements(FileSystem.ls)
    def ls(self, path, detail=False):
        path = _stringify_path(path)
        if detail:
            return self.showdir(path)
        return self.listdir(path)

    def walk(self, path):
        return self.fs.walk(_stringify_path(path))



# flake8: noqa

# Omniduct's automatic registration of Duct protocols requires that the subclass
# implementation be loaded into memory. Any protocol that should be enabled by
# default should be imported here.

from .caches.filesystem import FileSystemCache
from .databases.druid import DruidClient
from .databases.exasol import ExasolClient
from .databases.hiveserver2 import HiveServer2Client
from .databases.neo4j import Neo4jClient
from .databases.presto import PrestoClient
from .databases.pyspark import PySparkClient
from .databases.sqlalchemy import SQLAlchemyClient
from .filesystems.local import LocalFsClient
from .filesystems.s3 import S3Client
from .filesystems.webhdfs import WebHdfsClient
from .remotes.ssh import SSHClient
from .remotes.ssh_paramiko import ParamikoSSHClient
from .restful.base import RestClient



import inspect
import sys

import decorator
import six
from future.utils import raise_with_traceback


def function_args_as_kwargs(func, *args, **kwargs):
    if six.PY3 and not hasattr(sys, 'pypy_version_info'):
        arguments = inspect.signature(func).parameters.keys()
    else:
        arguments = inspect.getargspec(func).args
    kwargs.update(dict(zip(list(arguments), args)))
    return kwargs


@decorator.decorator
def require_connection(f, self, *args, **kwargs):
    """
    A wrapper to allow restoring of connection status in the event that
    connection issues result in failures. If so, we will attempt to retry the
    failed function call once more.
    """
    if not self._Duct__connected:
        self.connect()

    try:
        return f(self, *args, **kwargs)
    except Exception as e:
        # Check to see if it is possible that we failed due to connection issues.
        # If so, try again once more. If we fail again, raise.
        # TODO: Explore adding a DuctConnectionError class and filter this
        # handling to errors of that class.
        if not self.is_connected():
            self.connect()
            return f(self, *args, **kwargs)
        raise_with_traceback(e)



import json

from future.moves.urllib.parse import urljoin
from interface_meta import quirk_docs, override

from omniduct.duct import Duct
from omniduct.utils.decorators import require_connection


class RestClientBase(Duct):
    """
    A simple wrapper around the `requests` library to simplify the creation of
    REST clients.

    This allows all the automatic features around port forwarding from remote
    hosts to be inherited. This client can be used directly, or inherited by
    subclasses which can add methods specific to any REST service; and
    internally use `request` and `request_json` methods to access various
    endpoints.

    Attributes:
        server_protocol (str): The protocol to use when connecting to the
            remote host (default: `'http'`).
        assume_json (bool): Assume that responses will be JSON
            (default: `False`).
        endpoint_prefix (str): The base_url path relative to the host at
            which the API is accessible (default: `''`).
    """

    DUCT_TYPE = Duct.Type.RESTFUL

    @quirk_docs('_init', mro=True)
    def __init__(self, server_protocol='http', assume_json=False, endpoint_prefix='', **kwargs):
        """
        Args:
            server_protocol (str): The protocol to use when connecting to the
                remote host (default: `'http'`).
            assume_json (bool): Assume that responses will be JSON when calling
                instances of this class (default: `False`).
            endpoint_prefix (str): The base_url path relative to the host at
                which the API is accessible (default: `''`).
            **kwargs (dict): Additional keyword arguments passed on to
                subclasses.
        """
        Duct.__init_with_kwargs__(self, kwargs, port=80)

        self.server_protocol = server_protocol
        self.assume_json = assume_json
        self.endpoint_prefix = endpoint_prefix

        self._init(**kwargs)

    def _init(self):
        pass

    def __call__(self, endpoint, method='get', **kwargs):
        if self.assume_json:
            return self.request_json(endpoint, method=method, **kwargs)
        return self.request(endpoint, method=method, **kwargs)

    @property
    def base_url(self):
        """str: The base url of the REST API."""
        url = urljoin('{}://{}:{}'.format(self.server_protocol, self.host, self.port or 80), self.endpoint_prefix)
        if not url.endswith('/'):
            url += '/'
        return url

    @require_connection
    def request(self, endpoint, method='get', **kwargs):
        """
        Request data from a nominated endpoint.

        Args:
            endpoint (str): The endpoint from which to receive data.
            method (str): The method to use when requestion this resource.
            **kwargs (dict): Additional arguments to pass through to
                `requests.request`.

        Returns:
            requests.Response: The response object associated with this request.
        """
        import requests
        url = urljoin(self.base_url, endpoint)
        return requests.request(method, url, **kwargs)

    def request_json(self, endpoint, method='get', **kwargs):
        """
        Request JSON data from a nominated endpoint.

        Args:
            endpoint (str): The endpoint from which to receive data.
            method (str): The method to use when requestion this resource.
            **kwargs (dict): Additional arguments to pass through to
                `requests.request`.

        Returns:
            list, dict: The representation of the JSON response from the server.
        """
        request = self.request(endpoint, method=method, **kwargs)
        if not request.status_code == 200:
            try:
                raise RuntimeError("Server responded with HTTP response code {}, with content: {}.".format(request.status_code, json.dumps(request.json())))
            except:
                raise RuntimeError("Server responded with HTTP response code {}, with content: {}.".format(request.status_code, request.content.decode('utf-8')))
        return request.json()

    @override
    def _connect(self):
        pass

    @override
    def _is_connected(self):
        return True

    @override
    def _disconnect(self):
        pass


class RestClient(RestClientBase):
    """
    A trivial implementation of `RestClientBase` for basic REST access.
    """
    PROTOCOLS = ['rest']



from pandas.io.sql import SQLTable, SQLDatabase


def to_sql(df, name, schema, con, index, if_exists, mode='default', **kwargs):
    """
    Override the default `pandas.to_sql` method to allow for insertion of
    multiple rows of data at once. This is derived from the upstream patch at
    https://github.com/pandas-dev/pandas/pull/21401, and can be deprecated
    once it is merged and released in a new version of `pandas`.
    """
    assert mode in ('default', 'multi'), 'unexpected `to_sql` mode {}'.format(mode)
    if mode == 'default':
        return df.to_sql(
            name=name, schema=schema, con=con, index=index, if_exists=if_exists, **kwargs
        )
    else:
        nrows = len(df)
        if nrows == 0:
            return

        chunksize = kwargs.get('chunksize', nrows)
        if chunksize == 0:
            raise ValueError('chunksize argument should be non-zero')
        chunks = int(nrows / chunksize) + 1

        pd_sql = SQLDatabase(con)
        pd_table = SQLTable(
            name, pd_sql, frame=df, index=index, if_exists=if_exists,
            index_label=kwargs.get('insert_label'), schema=schema, dtype=kwargs.get('dtype')
        )
        pd_table.create()
        keys, data_list = pd_table.insert_data()

        with pd_sql.run_transaction() as conn:
            for i in range(chunks):
                start_i = i * chunksize
                end_i = min((i + 1) * chunksize, nrows)
                if start_i >= end_i:
                    break

                chunk_iter = zip(*[arr[start_i:end_i] for arr in data_list])
                data = [{k: v for k, v in zip(keys, row)} for row in chunk_iter]
                conn.execute(pd_table.table.insert(data))  # multivalues insert



import pickle
from distutils.version import LooseVersion

import pandas


class Serializer(object):

    @property
    def file_extension(self):
        return ""

    def serialize(self, obj, fh):
        raise NotImplementedError

    def deserialize(self, fh):
        raise NotImplementedError


class BytesSerializer(Serializer):

    @property
    def file_extension(self):
        return ".bytes"

    def serialize(self, obj, fh):
        assert isinstance(obj, bytes), "BytesSerializer requires incoming data be already encoded into a bytestring."
        fh.write(obj)

    def deserialize(self, fh):
        return fh.read()


class PickleSerializer(Serializer):

    @property
    def file_extension(self):
        return ".pickle"

    def serialize(self, obj, fh):
        return pickle.dump(obj, fh)

    def deserialize(self, fh):
        return pickle.load(fh)


class PandasSerializer(Serializer):

    @property
    def file_extension(self):
        return ".pandas"

    @classmethod
    def serialize(cls, formatted_data, fh):
        # compat: if pandas is old, to_pickle does not accept file handles
        if LooseVersion(pandas.__version__) <= LooseVersion('0.20.3'):
            fh.close()
            fh = fh.name
        return pandas.to_pickle(formatted_data, fh, compression=None)

    @classmethod
    def deserialize(cls, fh):
        return pandas.read_pickle(fh, compression=None)



import six
import yaml
from interface_meta import override

from omniduct.filesystems.base import FileSystemClient
from omniduct.filesystems.local import LocalFsClient

from .base import Cache


class FileSystemCache(Cache):
    """
    An implementation of `Cache` that wraps around a `FilesystemClient`.
    """

    PROTOCOLS = ['filesystem_cache']

    @override
    def _init(self, path, fs=None):
        """
        path (str): The top-level path of the cache in the filesystem.
        fs (FileSystemClient, str): The filesystem client to use as the
            datastore of this cache. If not specified, this will default to the
            local filesystem using `LocalFsClient`. If specified as a string,
            and connected to a `DuctRegistry`, upon first use an attempt will be
            made to look up a `FileSystemClient` instance in the registry by
            this name.
        """
        self.fs = fs or LocalFsClient()
        self.path = path
        # Currently config is not used, but will be in future versions
        self._config = None
        self.connection_fields += ('fs',)

    @override
    def _prepare(self):
        Cache._prepare(self)

        if self.registry is not None:
            if isinstance(self.fs, six.string_types):
                self.fs = self.registry.lookup(self.fs, kind=FileSystemCache.Type.FILESYSTEM)
        assert isinstance(self.fs, FileSystemClient), "Provided cache is not an instance of `omniduct.filesystems.base.FileSystemClient`."

        self._prepare_cache()

    def _prepare_cache(self):
        config_path = self.fs.path_join(self.path, 'config')
        if self.fs.exists(config_path):
            with self.fs.open(config_path) as fh:
                try:
                    return yaml.safe_load(fh)
                except yaml.error.YAMLError:
                    raise RuntimeError(
                        "Path nominated for cache ('{}') has a corrupt "
                        "configuration. Please manually empty or delete this "
                        "path cache, and try again.".format(self.path)
                    )

        # Cache needs initialising
        if self.fs.exists(self.path):
            if not self.fs.isdir(self.path):
                raise RuntimeError(
                    "Path nominated for cache ('{}') is not a directory.".format(self.path)
                )
            elif self.fs.listdir(self.path):
                raise RuntimeError(
                    "Cache directory ({}) needs to be initialised, and is not "
                    "empty. Please manually delete and/or empty this path, and "
                    "try again.".format(self.path)
                )
        else:  # Create cache directory
            self.fs.mkdir(self.path, recursive=True, exist_ok=True)

        # Write config file to mark cache as initialised
        with self.fs.open(config_path, 'w') as fh:
            yaml.safe_dump({'version': 1}, fh, default_flow_style=False)
        return {'version': 1}

    @override
    def _connect(self):
        self.fs.connect()

    @override
    def _is_connected(self):
        return self.fs.is_connected()

    @override
    def _disconnect(self):
        return self.fs.disconnect()

    # Implementations for abstract methods in Cache
    @override
    def _namespace(self, namespace):
        if namespace is None:
            return '__default__'
        assert isinstance(namespace, str) and namespace != 'config'
        return namespace

    @override
    def _get_namespaces(self):
        return [d for d in self.fs.listdir(self.path) if d != 'config']

    @override
    def _has_namespace(self, namespace):
        return self.fs.exists(self.fs.path_join(self.path, namespace))

    @override
    def _remove_namespace(self, namespace):
        return self.fs.remove(self.fs.path_join(self.path, namespace), recursive=True)

    @override
    def _get_keys(self, namespace):
        return self.fs.listdir(self.fs.path_join(self.path, namespace))

    @override
    def _has_key(self, namespace, key):
        return self.fs.exists(self.fs.path_join(self.path, namespace, key))

    @override
    def _remove_key(self, namespace, key):
        return self.fs.remove(self.fs.path_join(self.path, namespace, key), recursive=True)

    @override
    def _get_bytecount_for_key(self, namespace, key):
        path = self.fs.path_join(self.path, namespace, key)
        return sum([
            f.bytes
            for f in self.fs.dir(path)
        ])

    @override
    def _get_stream_for_key(self, namespace, key, stream_name, mode, create):
        path = self.fs.path_join(self.path, namespace, key)

        if create:
            self.fs.mkdir(path, recursive=True, exist_ok=True)

        return self.fs.open(self.fs.path_join(path, stream_name), mode=mode)



import base64
import textwrap

import jinja2
import six.moves.urllib as urllib

import omniduct


ABOUT_TEMPLATE_HTML = """
{%- if logo %}
<div><img src="{{logo}}" style="height: 200px;"/></div>
{%- endif %}
<b>{{name}}</b> v{{version}}<br/>

{%- if maintainers %}
<b>Maintainers:</b> {% for name, email in maintainers.items() %}{% if loop.index0 > 0 %}, {% endif %}<a href="mailto:{{email}}">{{name}}</a>{% endfor %}<br/>
{%- endif %}
{%- for name, value in attributes.items() %}
<b>{{name}}:</b> {% if value.startswith('http://') or value.startswith('https://') %}<a href='{{value}}'>{{value}}</a>{% else %}{{value}}{%endif%}<br />
{%- endfor %}
<b>Description:</b><br/>
{{description.strip()|replace('\n\n','<br/><br/>\n')}}
{%- if endorsements %}<br />
<br />
<b>Built upon:</b>
{%- for endorsement in endorsements %}
<div>
{%- if endorsement['logo'] -%}
<img src="{{endorsement['logo']}}" style="display: inline-block; height: 2em; width: 2em; object-size: contain; vertical-align: middle; margin-right: 0.5em;"/>
{%- else -%}
<span style='display: inline-block; width: 2em; height: 2em; vertical-align: middle; margin-right: 0.5em;'></span>
{%- endif -%}
{{endorsement['name']}} v{{endorsement['version'].strip()}}</div>
{%- endfor %}
{%- endif %}
"""

ABOUT_TEMPLATE_TEXT = """
{{name}}{% if version %} v{{version}}{% endif %}

{%- if maintainers %}
- Maintainers: {% for name, email in maintainers.items() %}{% if loop.index0 > 0 %}, {% endif %}{{name}} <{{email}}>{% endfor %}
{%- endif %}
{%- for name, value in attributes.items() %}
- {{name}}: {{value}}
{%- endfor %}

{{description}}
{% if endorsements %}
Built upon:
{%- for endorsement in endorsements %}
- {{endorsement['name']}} v{{endorsement['version'].strip()}}
{%- endfor %}
{%- endif %}
""".strip()


def show_about(name, version=None, logo=None, maintainers=None, attributes=None,
               description=None, endorsements=None, endorse_omniduct=True):
    """
    Output information about a project in HTML for notebooks and text otherwise.

    Args:
        name (str): The name of the project.
        version (str, None): The version of the project.
        logo (str, None): A local or remote uri for the project logo.
        maintainers (dict, None): A dictionary mapping name to email address for
            maintainers. If order is important pass an `OrderedDict`.
        attributes (dict, None): A dictionary mapping of project attributes to
            values. If order is important pass an `OrderedDict`.
        description (str, None): A description of the project.
        endorsements (list<dict>): A list of dependencies to highlight encoded
            as dictionaries of form: {'name': ..., 'version': ..., 'logo': ...}.
            Only `name` is required, and endorsements will be sorted by name.
        endorse_omniduct (bool): Whether to include omniduct in the list of
            endorsements (default: True).
    """

    endorsements = endorsements or []
    if endorse_omniduct:
        endorsements.append({
            'name': 'Omniduct',
            'version': omniduct.__version__,
            'logo': omniduct.__logo__
        })
    for endorsement in endorsements:
        endorsement['logo'] = get_image_url(endorsement.get('logo'))
    endorsements = sorted(endorsements, key=lambda x: x['name'])

    context = {
        'name': name,
        'version': version,
        'logo': get_image_url(logo),
        'maintainers': maintainers or {},
        'attributes': attributes or {},
        'description': textwrap.dedent(description).strip() if description else None,
        'endorsements': endorsements
    }

    try:
        from IPython import get_ipython
        from IPython.display import display, HTML
        ip = get_ipython()
        if ip is not None and ip.has_trait('kernel'):
            return display(HTML(jinja2.Template(ABOUT_TEMPLATE_HTML).render(**context)))
    except:
        pass

    # Textual fallback if HTML not running in a notebook
    print(textwrap.dedent(jinja2.Template(ABOUT_TEMPLATE_TEXT).render(**context)))


def get_image_url(uri):
    """
    Get a base64 URI if uri is a local path or pass through value otherwise.

    This is used to allow rendering of images in notebooks for local files,
    such as project logos.

    Args:
        uri (str, None): The local path of a logo or remote image.

    Returns:
        str: The uri of the image suitable for rendering in a notebook.
    """
    if not uri:
        return
    parsed = urllib.parse.urlparse(uri)
    if parsed.scheme in ('', 'file'):
        with open(parsed.path, 'rb') as image:
            return "data:image/png;base64,{}".format(base64.b64encode(image.read()).decode())
    return uri



# Connection Errors
class DuctAuthenticationError(RuntimeError):
    pass


class DuctConnectionError(RuntimeError):
    pass


class DuctServerUnreachable(RuntimeError):
    pass


# Lookups
class DuctNotFound(RuntimeError):
    pass


class DuctProtocolUnknown(RuntimeError):
    pass



from __future__ import absolute_import

from interface_meta import override

from omniduct.utils.debug import logger

from .base import DatabaseClient


class DruidClient(DatabaseClient):
    """
    This Duct connects to a Druid server using the `pydruid` python library.
    """

    PROTOCOLS = ['druid']
    DEFAULT_PORT = 80
    NAMESPACE_NAMES = ['table']
    NAMESPACE_QUOTECHAR = '"'
    NAMESPACE_SEPARATOR = '.'

    @override
    def _init(self):
        self.__druid = None

    # Connection
    @override
    def _connect(self):
        from pydruid.db import connect
        logger.info('Connecting to Druid database ...')
        self.__druid = connect(self.host, self.port, path='/druid/v2/sql/', scheme='http')
        if self.username or self.password:
            logger.warning(
                'Duct username and password not passed to pydruid connection. '
                'pydruid connection currently does not allow these fields to be passed.'
            )

    @override
    def _is_connected(self):
        return self.__druid is not None

    @override
    def _disconnect(self):
        logger.info('Disconnecting from Druid database ...')
        try:
            self.__druid.close()
        except Exception:
            pass
        self.__druid = None

    # Querying
    @override
    def _execute(self, statement, cursor, wait, session_properties):
        cursor = cursor or self.__druid.cursor()
        cursor.execute(statement)
        return cursor

    @override
    def _table_list(self, namespace, like=None, **kwargs):
        cmd = "SELECT * FROM INFORMATION_SCHEMA.TABLES"
        return self.query(cmd, **kwargs)

    @override
    def _table_exists(self, table, **kwargs):
        logger.disabled = True
        try:
            self.table_desc(table, **kwargs)
            return True
        except:
            return False
        finally:
            logger.disabled = False

    @override
    def _table_drop(self, table, **kwargs):
        raise NotImplementedError

    @override
    def _table_desc(self, table, **kwargs):
        query = ("""
            SELECT
                TABLE_SCHEMA
                , TABLE_NAME
                , COLUMN_NAME
                , ORDINAL_POSITION
                , COLUMN_DEFAULT
                , IS_NULLABLE
                , DATA_TYPE
            FROM INFORMATION_SCHEMA.COLUMNS
            WHERE TABLE_NAME = '{}'""").format(table)
        return self.query(query, **kwargs)

    @override
    def _table_head(self, table, n=10, **kwargs):
        return self.query("SELECT * FROM {} LIMIT {}".format(table, n), **kwargs)

    @override
    def _table_props(self, table, **kwargs):
        raise NotImplementedError



from __future__ import absolute_import

from interface_meta import override

from omniduct.utils.debug import logger

from .base import DatabaseClient


class ExasolClient(DatabaseClient):
    """
    This client connects to an Exasol service using the `pyexasol` python library.

    Example Config
    --------------

    databases:
        exasol_db:
            protocol: exasol
            host:
              - 'localhost:8563'
              - 'localhost:8564'
            username: exasol_user
            password: ****
            schema: users
    """

    PROTOCOLS = ["exasol"]
    DEFAULT_PORT = 8563
    NAMESPACE_NAMES = ["schema", "table"]
    NAMESPACE_QUOTECHAR = '"'
    NAMESPACE_SEPARATOR = "."

    @override
    @property
    def NAMESPACE_DEFAULT(self):
        return {"schema": self.schema}

    @override
    def _init(self, schema=None, engine_opts=None):
        self.__exasol = None

        self.schema = schema
        self.connection_fields += ("schema",)
        self.engine_opts = engine_opts or {}

    @override
    def _connect(self):
        import pyexasol

        logger.info("Connecting to Exasol ...")
        self.__exasol = pyexasol.connect(
            dsn="{host}:{port}".format(host=self.host, port=self.port),
            user=self.username,
            password=self.password,
            **self.engine_opts
        )

    @override
    def _is_connected(self):
        return self.__exasol is not None

    @override
    def _disconnect(self):
        try:
            self.__exasol.close()
        except Exception:
            pass
        self.__exasol = None

    @override
    def _execute(self, statement, cursor, wait, session_properties, query=True):
        # pyexasol.ExaStatement has a similar interface to that of
        # a DBAPI2 cursor.
        cursor = cursor or self.__exasol.execute(statement)

        # hacky: make the result look like a cursor.
        # cursor.columns returns a dict with the required attributes
        cursor.description = []
        for key, values in cursor.columns().items():
            cursor.description.append(
                (
                    key,
                    values.get("type", None),
                    values.get("size", None),
                    values.get("size", None),
                    values.get("precision", None),
                    values.get("scale", None),
                    True,
                )
            )

        return cursor

    @override
    def _query_to_table(self, statement, table, if_exists, **kwargs):
        statements = []

        if if_exists == "fail" and self.table_exists(table):
            raise RuntimeError("Table {} already exists!".format(table))
        elif if_exists == "replace":
            statements.append("DROP TABLE IF EXISTS {};".format(table))
        elif if_exists == "append":
            raise NotImplementedError(
                "Append operations have not been implemented for {}.".format(
                    self.__class__.__name__
                )
            )

        statement = "CREATE TABLE {table} AS ({statement})".format(
            table=table, statement=statement
        )
        return self.execute(statement, **kwargs)

    @override
    def _table_list(self, namespace, **kwargs):
        # Since this namespace is a conditional, exasol requires single quotations
        # instead of double quotations. " -> '
        query = (
            "SELECT TABLE_NAME FROM EXA_ALL_TABLES WHERE table_schema={}"
            .format(namespace.render(quote_char="'"))
        )
        return self.query(query, **kwargs)

    @override
    def _table_exists(self, table, **kwargs):
        logger.disabled = True
        try:
            self.table_desc(table, **kwargs)
            return True
        except:
            return False
        finally:
            logger.disabled = False

    @override
    def _table_drop(self, table, **kwargs):
        # Schema and tables are always under uppercase namespaces.
        return self.execute(
            "DROP TABLE {table}".format(table=str(table).upper()),
            **kwargs
        )

    @override
    def _table_desc(self, table, **kwargs):
        # Schema and tables are always under uppercase namespaces.
        return self.query(
            "DESCRIBE {0}".format(str(table).upper()),
            **kwargs
        )

    @override
    def _table_head(self, table, n=10, **kwargs):
        # Schema and tables are always under uppercase namespaces.
        return self.query(
            "SELECT * FROM {} LIMIT {}".format(str(table).upper(), n),
            **kwargs
        )

    @override
    def _table_props(self, table, **kwargs):
        raise NotImplementedError



import io
from abc import abstractmethod
from collections import OrderedDict, namedtuple

import pandas as pd
from interface_meta import quirk_docs, override

from omniduct.duct import Duct
from omniduct.utils.decorators import require_connection
from omniduct.utils.magics import MagicsProvider, process_line_arguments


class FileSystemClient(Duct, MagicsProvider):
    """
    An abstract class providing the common API for all filesystem clients.

    Class Attributes:
        DUCT_TYPE (`Duct.Type`): The type of `Duct` protocol implemented by this class.
        DEFAULT_PORT (int): The default port for the filesystem service (defined
            by subclasses).
    """

    DUCT_TYPE = Duct.Type.FILESYSTEM
    DEFAULT_PORT = None

    @quirk_docs('_init', mro=True)
    def __init__(self, cwd=None, home=None, read_only=False, global_writes=False, **kwargs):
        """
        cwd (None, str): The path prefix to use as the current working directory
            (if None, the user's home directory is used where that makes sense).
        home (None, str): The path prefix to use as the current users' home
            directory. If not specified, it will default to an implementation-
            specific value (often '/').
        read_only (bool): Whether the filesystem should only be able to perform
            read operations.
        global_writes (bool): Whether to allow writes outside of the user's home
            folder.
        **kwargs (dict): Additional keyword arguments to passed on to subclasses.
        """
        Duct.__init_with_kwargs__(self, kwargs, port=self.DEFAULT_PORT)
        self._path_cwd = cwd
        self.__path_home = home
        self.read_only = read_only
        self.global_writes = global_writes
        self._init(**kwargs)

    @abstractmethod
    def _init(self):
        pass

    # Path properties and helpers

    @property
    @quirk_docs('_path_home')
    @require_connection
    def path_home(self):
        """
        str: The path prefix to use as the current users' home directory. Unless
        `cwd` is set, this will be the prefix to use for all non-absolute path
        references on this filesystem. This is assumed not to change between
        connections, and so will not be updated on client reconnections. Unless
        `global_writes` is set to `True`, this will be the only folder into
        which this client is permitted to write.
        """
        if not self.__path_home:
            self.__path_home = self._path_home()
        return self.__path_home

    @path_home.setter
    def path_home(self, path_home):
        if path_home is not None and not path_home.startswith(self.path_separator):
            raise ValueError("The home path must be absolute. Received: '{}'.".format(path_home))
        self.__path_home = path_home

    @abstractmethod
    def _path_home(self):
        return NotImplementedError

    @property
    def path_cwd(self):
        """
        str: The path prefix associated with the current working directory. If
        not otherwise set, it will be the users' home directory, and will be the
        prefix used by all non-absolute path references on this filesystem.
        """
        return self._path_cwd or self.path_home

    @path_cwd.setter
    def path_cwd(self, path_cwd):
        path_cwd = self._path(path_cwd)
        assert self.isdir(path_cwd), "Specified path does not exist."
        self._path_cwd = path_cwd

    @property
    @quirk_docs('_path_separator')
    def path_separator(self):
        """
        str: The character(s) to use in separating path components. Typically
        this will be '/'.
        """
        return self._path_separator()

    @abstractmethod
    def _path_separator(self):
        raise NotImplementedError

    def path_join(self, path, *components):
        """
        Generate a new path by joining together multiple paths.

        If any component starts with `self.path_separator` or '~', then all
        previous path components are discarded, and the effective base path
        becomes that component (with '~' expanding to `self.path_home`). Note
        that this method does *not* simplify paths components like '..'. Use
        `self.path_normpath` for this purpose.

        Args:
            path (str): The base path to which components should be joined.
            *components (str): Any additional components to join to the base
                path.

        Returns:
            str: The path resulting from joining all of the components nominated,
            in order, to the base path.
        """
        for component in components:
            if component.startswith('~'):
                path = self.path_home + component[1:]
            elif component.startswith(self.path_separator):
                path = component
            else:
                path = '{}{}{}'.format(path, self.path_separator if not path.endswith(self.path_separator) else '', component)
        return path

    def path_basename(self, path):
        """
        Extract the last component of a given path.

        Components are determined by splitting by `self.path_separator`.
        Note that if a path ends with a path separator, the basename will be
        the empty string.

        Args:
            path (str): The path from which the basename should be extracted.

        Returns:
            str: The extracted basename.
        """
        return self._path(path).split(self.path_separator)[-1]

    def path_dirname(self, path):
        """
        Extract the parent directory for provided path.

        This method returns the entire path except for the basename (the last
        component), where components are determined by splitting by
        `self.path_separator`.

        Args:
            path (str): The path from which the directory path should be
                extracted.

        Returns:
            str: The extracted directory path.
        """
        return self.path_separator.join(self._path(path).split(self.path_separator)[:-1])

    def path_normpath(self, path):
        """
        Normalise a pathname.

        This method returns the normalised (absolute) path corresponding to `path`
        on this filesystem.

        Args:
            path (str): The path to normalise (make absolute).

        Returns:
            str: The normalised path.
        """
        components = self._path(path).split(self.path_separator)
        out_path = []
        for component in components:
            if component == '' and len(out_path) > 0:
                continue
            if component == '.':
                continue
            elif component == '..':
                if len(out_path) > 1:
                    out_path.pop()
                else:
                    raise RuntimeError("Cannot access parent directory of filesystem root.")
            else:
                out_path.append(component)
        if len(out_path) == 1 and out_path[0] == '':
            return self.path_separator
        return self.path_separator.join(out_path)

    def _path(self, path=None):
        return self.path_cwd if path is None else self.path_join(self.path_cwd, path)

    def _path_in_home_dir(self, path):
        return self.path_normpath(path).startswith(self.path_home)

    @property
    def read_only(self):
        """
        bool: Whether this filesystem client should be permitted to attempt any
        write operations.
        """
        return self._read_only

    @read_only.setter
    def read_only(self, read_only):
        self._read_only = read_only

    @property
    def global_writes(self):
        """
        bool: Whether writes should be permitted outside of home directory. This
        write-lock is designed to prevent inadvertent scripted writing in
        potentially dangerous places.
        """
        return self._global_writes

    @global_writes.setter
    def global_writes(self, global_writes):
        self._global_writes = global_writes

    def _assert_path_is_writable(self, path):
        if self.read_only:
            raise RuntimeError("This filesystem client is configured for read-only access. Set `{}`.`read_only` to `False` to override.".format(self.name))
        if not self.global_writes and not self._path_in_home_dir(path):
            raise RuntimeError("Attempt to write outside of home directory without setting `{}`.`global_writes` to `True`.".format(self.name))
        return True

    # Filesystem accessors

    @quirk_docs('_exists')
    @require_connection
    def exists(self, path):
        """
        Check whether nominated path exists on this filesytem.

        Args:
            path (str): The path for which to check existence.

        Returns:
            bool: `True` if file/folder exists at nominated path, and `False`
                otherwise.
        """
        return self._exists(self._path(path))

    @abstractmethod
    def _exists(self, path):
        raise NotImplementedError

    @quirk_docs('_isdir')
    @require_connection
    def isdir(self, path):
        """
        Check whether a nominated path is directory.

        Args:
            path (str): The path for which to check directory nature.

        Returns:
            bool: `True` if folder exists at nominated path, and `False`
            otherwise.
        """
        return self._isdir(self._path(path))

    @abstractmethod
    def _isdir(self, path):
        raise NotImplementedError

    @quirk_docs('_isfile')
    @require_connection
    def isfile(self, path):
        """
        Check whether a nominated path is a file.

        Args:
            path (str): The path for which to check file nature.

        Returns:
            bool: `True` if a file exists at nominated path, and `False`
            otherwise.
        """
        return self._isfile(self._path(path))

    def _isfile(self, path):
        return not self._isdir(path)

    # Directory handling

    @abstractmethod
    def _dir(self, path):
        """
        This method should return a generator over `FileSystemFileDesc` objects.
        """
        raise NotImplementedError

    @quirk_docs('_dir')
    @require_connection
    def dir(self, path=None):
        """
        Retrieve information about the children of a nominated directory.

        This method returns a generator over `FileSystemFileDesc` objects that
        represent the files/directories that a present as children of the
        nominated path. If `path` is not a directory, an exception is raised.
        The path is interpreted as being relative to the current working
        directory (on remote filesytems, this will typically be the home
        folder).

        Args:
            path (str): The path to examine for children.

        Returns:
            generator<FileSystemFileDesc>: The children of `path` represented as
            `FileSystemFileDesc` objects.
        """
        assert self.isdir(path), "'{}' is not a valid directory.".format(path)
        return self._dir(self._path(path))

    def listdir(self, path=None):
        """
        Retrieve the names of the children of a nomianted directory.

        This method inspects the contents of a directory using `.dir(path)`, and
        returns the names of child members as strings. `path` is interpreted
        relative to the current working directory (on remote filesytems, this
        will typically be the home folder).

        Args:
            path (str): The path of the directory from which to enumerate filenames.

        Returns:
            list<str>: The names of all children of the nominated directory.
        """
        return [f.name for f in self.dir(self._path(path))]

    @require_connection
    def showdir(self, path=None):
        """
        Return a dataframe representation of a directory.

        This method returns a `pandas.DataFrame` representation of the contents of
        a path, which are retrieved using `.dir(path)`. The exact columns will
        vary from filesystem to filesystem, depending on the fields returned
        by `.dir()`, but the returned DataFrame is guaranteed to at least have
        the columns: 'name' and 'type'.

        Args:
            path (str): The path of the directory from which to show contents.

        Returns:
            pandas.DataFrame: A DataFrame representation of the contents of the
            nominated directory.
        """
        assert self.isdir(path), "'{}' is not a valid directory.".format(path)
        return self._showdir(self._path(path))

    def _showdir(self, path):
        data = [f.as_dict() for f in self._dir(path)]
        if len(data) > 0:
            return (
                pd.DataFrame(data)
                .sort_values(['type', 'name'])
                .reset_index(drop=True)
                .dropna(axis='columns', how='all')
                .drop(axis=1, labels=['fs', 'path'])
            )
        else:
            return "Directory has no contents."

    @quirk_docs('_walk')
    @require_connection
    def walk(self, path=None):
        """
        Explore the filesystem tree starting at a nominated path.

        This method returns a generator which recursively walks over all paths
        that are children of `path`, one result for each directory, of form:
        (<path name>, [<directory 1>, ...], [<file 1>, ...])

        Args:
            path (str): The path of the directory from which to enumerate
                contents.

        Returns:
            generator<tuple>: A generator of tuples, each tuple being associated
            with one directory that is either `path` or one of its descendants.
        """
        assert self.isdir(path), "'{}' is not a valid directory.".format(path)
        return self._walk(self._path(path))

    def _walk(self, path):
        dirs = []
        files = []
        for f in self._dir(path):
            if f.type == 'directory':
                dirs.append(f.name)
            else:
                files.append(f.name)
        yield (path, dirs, files)

        for dir in dirs:
            for walked in self._walk(self._path(self.path_join(path, dir))):  # Note: using _walk directly here, which may fail if disconnected during walk.
                yield walked

    @quirk_docs('_find')
    @require_connection
    def find(self, path_prefix=None, **attrs):
        """
        Find a file or directory based on certain attributes.

        This method searches for files or folders which satisfy certain
        constraints on the attributes of the file (as encoded into
        `FileSystemFileDesc`). Note that without attribute constraints,
        this method will function identically to `self.dir`.

        Args:
            path_prefix (str): The path under which files/directories should be
                found.
            **attrs (dict): Constraints on the fields of the `FileSystemFileDesc`
                objects associated with this filesystem, as constant values or
                callable objects (in which case the object will be called and
                should return True if attribute value is match, and False
                otherwise).

        Returns:
            generator<FileSystemFileDesc>: A generator over `FileSystemFileDesc`
                objects that are descendents of `path_prefix` and which statisfy
                provided constraints.
        """
        assert self.isdir(path_prefix), "'{0}' is not a valid directory. Did you mean `.find(name='{0}')`?".format(path_prefix)
        return self._find(self._path(path_prefix), **attrs)

    def _find(self, path_prefix, **attrs):

        def is_match(f):
            for attr, value in attrs.items():
                if hasattr(value, '__call__') and not value(f.as_dict().get(attr)):
                    return False
                elif value != f.as_dict().get(attr):
                    return False
            return True

        dirs = []
        for f in self._dir(path_prefix):
            if f.type == 'directory':
                dirs.append(f.name)
            if is_match(f):
                yield f

        for dir in dirs:
            for match in self._find(self._path(self.path_join(path_prefix, dir)), **attrs):  # Note: using _find directly here, which may fail if disconnected during find.
                yield match

    @quirk_docs('_mkdir')
    @require_connection
    def mkdir(self, path, recursive=True, exist_ok=False):
        """
        Create a directory at the given path.

        Args:
            path (str): The path of the directory to create.
            recursive (bool): Whether to recursively create any parents of this
                path if they do not already exist.

        Note: `exist_ok` is passed onto subclass implementations of `_mkdir`
        rather that implementing the existence check using `.exists` so that
        they can avoid the overhead associated with multiple operations, which
        can be costly in some cases.
        """
        self._assert_path_is_writable(path)
        return self._mkdir(self._path(path), recursive, exist_ok)

    @abstractmethod
    def _mkdir(self, path, recursive, exist_ok):
        raise NotImplementedError

    @quirk_docs('_remove')
    @require_connection
    def remove(self, path, recursive=False):
        """
        Remove file(s) at a nominated path.

        Directories (and their contents) will not be removed unless `recursive`
        is set to `True`.

        Args:
            path (str): The path of the file/directory to be removed.
            recursive (bool): Whether to remove directories and all of their
                contents.
        """
        self._assert_path_is_writable(path)
        if not self.exists(path):
            raise IOError("No file(s) exist at path '{}'.".format(path))
        if self.isdir(path) and not recursive:
            raise IOError("Attempt to remove directory '{}' without passing `recursive=True`.".format(path))
        return self._remove(self._path(path), recursive)

    @abstractmethod
    def _remove(self, path, recursive):
        raise NotImplementedError

    # File handling

    @quirk_docs('_open')
    @require_connection
    def open(self, path, mode='rt'):
        """
        Open a file for reading and/or writing.

        This method opens the file at the given path for reading and/or writing
        operations. The object returned is programmatically interchangeable with
        any other Python file-like object, including specification of file
        modes. If the file is opened in write mode, changes will only be flushed
        to the source filesystem when the file is closed.

        Args:
            path (str): The path of the file to open.
            mode (str): All standard Python file modes.

        Returns:
            FileSystemFile or file-like: An opened file-like object.
        """
        if 'w' in mode or 'a' in mode or '+' in mode:
            self._assert_path_is_writable(path)
        return self._open(self._path(path), mode=mode)

    def _open(self, path, mode):
        return FileSystemFile(self, path, mode)

    @quirk_docs('_file_read_')
    @require_connection
    def _file_read(self, path, size=-1, offset=0, binary=False):
        """
        This method is used by `FileSystemFile` to read the contents of files.
        `._file_read_` may be left unimplemented if `.open()` returns a different
        kind of file handle.

        Args:
            path (str): The path of the file to be read.
            size (int): The number of bytes to read at a time (-1 for max possible).
            offset (int): The offset in bytes from the start of the file.
            binary (bool): Whether to read the file in binary mode.

        Returns:
            str or bytes: The contents of the file.
        """
        return self._file_read_(self._path(path), size=size, offset=offset, binary=binary)

    def _file_read_(self, path, size=-1, offset=0, binary=False):
        raise NotImplementedError

    @quirk_docs('_file_write_')
    @require_connection
    def _file_write(self, path, s, binary=False):
        """
        This method is used by `FileSystemFile` to write to files.
        `._file_write_` may be left unimplemented if `.open()` returns a different
        kind of file handle.

        Args:
            path (str): The path of the file to be read.
            s (str, bytes): The content to be written to the file.
            binary (bool): Whether to read the file in binary mode.

        Returns:
            int: Number of bytes/characters written.
        """
        self._assert_path_is_writable(path)
        return self._file_write_(self._path(path), s, binary)

    def _file_write_(self, path, s, binary):
        raise NotImplementedError

    @quirk_docs('_file_append_')
    @require_connection
    def _file_append(self, path, s, binary=False):
        """
        This method is used by `FileSystemFile` to append content to files.
        `._file_append_` may be left unimplemented if `.open()` returns a different
        kind of file handle.

        Args:
            path (str): The path of the file to be read.
            s (str, bytes): The content to be appended to the file.
            binary (bool): Whether to read the file in binary mode.

        Returns:
            int: Number of bytes/characters written.
        """
        self._assert_path_is_writable(path)
        return self._file_append_(self._path(path), s, binary)

    def _file_append_(self, path, s, binary):
        raise NotImplementedError

    # File transfer

    @quirk_docs('_download')
    def download(self, source, dest=None, overwrite=False, fs=None):
        """
        Download files to another filesystem.

        This method (recursively) downloads a file/folder from path `source` on
        this filesystem to the path `dest` on filesytem `fs`, overwriting any
        existing file if `overwrite` is `True`.

        Args:
            source (str): The path on this filesystem of the file to download to
                the nominated filesystem (`fs`). If `source` ends
                with '/' then contents of the the `source` directory will be
                copied into destination folder, and will throw an error if path
                does not resolve to a directory.
            dest (str): The destination path on filesystem (`fs`). If not
                specified, the file/folder is downloaded into the default path,
                usually one's home folder. If `dest` ends with '/',
                and corresponds to a directory, the contents of source will be
                copied instead of copying the entire folder. If `dest` is
                otherwise a directory, an exception will be raised.
            overwrite (bool): `True` if the contents of any existing file by the
                same name should be overwritten, `False` otherwise.
            fs (FileSystemClient): The FileSystemClient into which the nominated
                file/folder `source` should be downloaded. If not specified,
                defaults to the local filesystem.
        """

        # TODO: Consider integration with `odo` for optimised data transfers.

        if fs is None:
            from .local import LocalFsClient
            fs = LocalFsClient()

        source = self._path(source)
        dest = fs._path(dest or self.path_basename(source))

        if dest.endswith(fs.path_separator):
            assert fs.isdir(dest), "No such directory `{}`".format(dest)
            if not source.endswith(self.path_separator):
                dest = fs.path_join(fs._path(dest), self.path_basename(source))

        # A mapping of source to dest paths on the respective filesystems
        # In format: (source, dest, isdir?)
        targets = []

        if self.isdir(source):
            target_prefix = (
                source if source.endswith(self.path_separator) else source + self.path_separator
            )
            targets.append((source, dest, True))

            for path, dirs, files in self.walk(source):
                for dir in dirs:
                    target_source = self.path_join(path, dir)
                    targets.append((
                        target_source,
                        fs.path_join(dest, *target_source[len(target_prefix):].split(self.path_separator)),
                        True
                    ))
                for file in files:
                    target_source = self.path_join(path, file)
                    targets.append((
                        target_source,
                        fs.path_join(dest, *target_source[len(target_prefix):].split(self.path_separator)),
                        False
                    ))
        else:
            targets.append((source, dest, False))

        for target in targets:
            if target[2] and not fs.isdir(target[1]):
                fs.mkdir(target[1], exist_ok=True)
            elif not target[2]:
                self._download(target[0], target[1], overwrite, fs)

    def _download(self, source, dest, overwrite, fs):
        if not overwrite and fs.exists(dest):
            raise RuntimeError("File already exists on filesystem.")
        with self.open(source, 'rb') as f_src:
            with fs.open(dest, 'wb') as f_dest:
                f_dest.write(f_src.read())

    def upload(self, source, dest=None, overwrite=False, fs=None):
        """
        Upload files from another filesystem.

        This method (recursively) uploads a file/folder from path `source` on
        filesystem `fs` to the path `dest` on this filesytem, overwriting any
        existing file if `overwrite` is `True`. This is equivalent to
        `fs.download(..., fs=self)`.

        Args:
            source (str): The path on the specified filesystem (`fs`) of the
                file to upload to this filesystem. If `source` ends with '/',
                and corresponds to a directory, the contents of source will be
                copied instead of copying the entire folder.
            dest (str): The destination path on this filesystem. If not
                specified, the file/folder is uploaded into the default path,
                usually one's home folder, on this filesystem. If `dest` ends
                with '/' then file will be copied into destination folder, and
                will throw an error if path does not resolve to a directory.
            overwrite (bool): `True` if the contents of any existing file by the
                same name should be overwritten, `False` otherwise.
            fs (FileSystemClient): The FileSystemClient from which to load the
                file/folder at `source`. If not specified, defaults to the local
                filesystem.
        """
        if fs is None:
            from .local import LocalFsClient
            fs = LocalFsClient()
        return fs.download(source, dest, overwrite, self)

    # Magics
    @override
    def _register_magics(self, base_name):
        from IPython.core.magic import register_line_magic, register_cell_magic

        @register_line_magic("{}.listdir".format(base_name))
        @process_line_arguments
        def listdir(path=''):
            return self.listdir(path)

        @register_line_magic("{}.showdir".format(base_name))
        @process_line_arguments
        def showdir(path=''):
            return self.showdir(path)

        @register_line_magic("{}.read".format(base_name))
        @process_line_arguments
        def read_file(path):
            with self.open(path) as f:
                return f.read()

        @register_cell_magic("{}.write".format(base_name))
        @process_line_arguments
        def write_file(cell, path):
            with self.open(path, 'w') as f:
                f.write(cell)

    # PyArrow compat
    @property
    def pyarrow_fs(self):
        from ._pyarrow_compat import OmniductFileSystem
        return OmniductFileSystem(self)


class FileSystemFile(object):
    """
    A file-like implementation that is interchangeable with native Python file
    objects, allowing remote files to be treated identically to local files
    both by omniduct, the user and other libraries.
    """

    def __init__(self, fs, path, mode='r'):
        self.fs = fs
        self.path = path
        self.mode = mode
        self.offset = 0
        self.closed = False
        self.__modified = False

        if self.binary_mode:
            self.__io_buffer = io.BytesIO()
        else:
            self.__io_buffer = io.StringIO()

        if 'w' not in self.mode:
            self.__io_buffer.write(self.fs._file_read(self.path, binary=self.binary_mode))
            if not self.appending:
                self.__io_buffer.seek(0)

    @property
    def name(self):
        return self.path

    @property
    def mode(self):
        return self.__mode

    @mode.setter
    def mode(self, mode):
        try:
            assert len(set(mode)) == len(mode)
            assert sum(l in mode for l in ['r', 'w', 'a', '+', 't', 'b']) == len(mode)
            assert sum(l in mode for l in ['r', 'w', 'a']) == 1
            assert sum(l in mode for l in ['t', 'b']) < 2
        except AssertionError:
            raise ValueError("invalid mode: '{}'".format(mode))
        self.__mode = mode

    @property
    def readable(self):
        return 'r' in self.mode or '+' in self.mode

    @property
    def writable(self):
        return 'w' in self.mode or 'a' in self.mode or '+' in self.mode

    @property
    def seekable(self):
        return True

    @property
    def appending(self):
        return 'a' in self.mode

    @property
    def binary_mode(self):
        return 'b' in self.mode

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.close()

    def close(self):
        self.flush()
        self.closed = True

    def __del__(self):
        self.close()

    def flush(self):
        if not self.writable or not self.__modified:
            return

        # For the time being, just write out entire buffer. We can consider something cleverer later.
        offset = self.__io_buffer.tell()
        self.__io_buffer.seek(0)
        self.fs._file_write(self.path, self.__io_buffer.read(), binary=self.binary_mode)
        self.__io_buffer.seek(offset)

        self.__modified = False

    def isatty(self):
        return self.__io_buffer.isatty()

    @property
    def newlines(self):
        return '\n'  # TODO: Support non-Unix newlines?

    def read(self, size=-1):
        if not self.readable:
            raise io.UnsupportedOperation("File not open for reading.")
        return self.__io_buffer.read(size)

    def readline(self, size=-1):
        if not self.readable:
            raise io.UnsupportedOperation("File not open for reading.")
        return self.__io_buffer.readline(size)

    def readlines(self, hint=-1):
        if not self.readable:
            raise io.UnsupportedOperation("File not open for reading.")
        return self.__io_buffer.readlines(hint)

    def seek(self, pos, whence=0):
        return self.__io_buffer.seek(pos, whence)

    def tell(self):
        return self.__io_buffer.tell()

    def write(self, s):
        if not self.writable:
            raise io.UnsupportedOperation("File not open for writing.")
        self.__io_buffer.write(s)
        self.__modified = True

    def __iter__(self):
        return self

    def __next__(self):
        line = self.readline()
        if not line:
            raise StopIteration
        return line

    next = __next__  # Python 2

    # Additional methods from BufferedIOBase for compatibility

    def read1(self, size=-1):
        return self.read(size)

    def detach(self):
        raise io.UnsupportedOperation()

    def readinto(self, buffer):
        data = self.read()
        buffer[:len(data)] = data
        return len(data)

    def readinto1(self, buffer):
        return self.readinto(buffer)


class FileSystemFileDesc(namedtuple('Node', [
    'fs',
    'path',
    'name',
    'type',
    'bytes',
    'owner',
    'group',
    'permissions',
    'created',
    'last_modified',
    'last_accessed',
    'extra',
])):
    """
    A representation of a file/directory stored within an Omniduct
    FileSystemClient.
    """

    __slots__ = ()

    def __new__(cls, fs, path, name, type, bytes=None, owner=None,
                group=None, permissions=None, created=None, last_modified=None,
                last_accessed=None, **extra):
        assert type in ('directory', 'file')
        return (
            super(FileSystemFileDesc, cls)
            .__new__(cls,
                     fs=fs,
                     path=path,
                     name=name,
                     type=type,
                     bytes=bytes,
                     owner=owner,
                     group=group,
                     permissions=permissions,
                     created=created,
                     last_modified=last_modified,
                     last_accessed=last_accessed,
                     extra=extra)
        )

    def as_dict(self):
        d = OrderedDict([
            ('fs', self.fs),
            ('path', self.path),
            ('type', self.type),
            ('name', self.name),
            ('bytes', self.bytes),
            ('owner', self.owner),
            ('group', self.group),
            ('permissions', self.permissions),
            ('created', self.created),
            ('last_modified', self.last_modified),
            ('last_accessed', self.last_accessed),
        ])
        d.update(self.extra)
        return d

    # Convenience methods

    def open(self, mode='rt'):
        assert self.type == 'file', "`.open(...)` is only appropriate for files."
        return self.fs.open(self.path, mode=mode)

    def dir(self):
        assert self.type == 'directory', "`.dir(...)` is only appropriate for directories."
        return self.fs.dir(self.path)

    def listdir(self):
        assert self.type == 'directory', "`.listdir(...)` is only appropriate for directories."
        return self.fs.listdir(self.path)

    def showdir(self):
        assert self.type == 'directory', "`.showdir(...)` is only appropriate for directories."
        return self.fs.showdir(self.path)

    def find(self, **attrs):
        assert self.type == 'directory', "`.find(...)` is only appropriate for directories."
        return self.fs.find(self.path, **attrs)

    def download(self, dest=None, overwrite=False, fs=None):
        return self.fs.download(self.path, dest=dest, overwrite=overwrite, fs=fs)



import logging

from interface_meta import override

from omniduct.filesystems.base import FileSystemClient, FileSystemFileDesc

# Python 2 compatibility imports
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError


class S3Client(FileSystemClient):
    """
    This Duct connects to an Amazon S3 bucket instance using the `boto3`
    library. Authentication is (optionally) handled using `opinel`.

    Attributes:
        bucket (str): The name of the Amazon S3 bucket to use.
        aws_profile (str): The name of configured AWS profile to use. This should
            refer to the name of a profile configured in, for example,
            `~/.aws/credentials`. Authentication is handled by the `opinel`
            library, which is also aware of environment variables.
    """

    PROTOCOLS = ['s3']
    DEFAULT_PORT = 80

    @override
    def _init(self, bucket=None, aws_profile=None, use_opinel=False,
              session=None, path_separator='/', skip_hadoop_artifacts=True):
        """
        bucket (str): The name of the Amazon S3 bucket to use.
        aws_profile (str): The name of configured AWS profile to use. This should
            refer to the name of a profile configured in, for example,
            `~/.aws/credentials`. Authentication is (optionally) handled by the
            `opinel` library, which is also aware of environment variables.
        use_opinel (bool): Use Opinel to extract AWS credentials. This is mainly
            useful if you have used opinel to set up MFA. Note: Opinel must be
            installed manually alongside omniduct to take advantage of this
            feature.
        session (botocore.session.Session): A pre-configured botocore Session
            instance to use instead of creating a new one when this client
            connects.
        path_separator (str): Amazon S3 is essentially a key-based storage
            system, and so one is free to choose an arbitrary "directory"
            separator. This defaults to '/' for consistency with other
            filesystems.
        skip_hadoop_artifacts (bool): Whether to skip hadoop artifacts like
            '*_$folder$' when enumerating directories (default=True).

        Note 1: aws_profile, if specified, should be the name of a profile as
        specified in ~/.aws/credentials. Authentication is handled by the
        `opinel` library, which is also aware of environment variables.
        Set up your command line aws client, and if it works, this should too.

        Note 2: Some institutions have nuanced AWS configurations that with
        configurations that are generated by scripts. It may be useful in these
        environments to subclass `S3Client` and override the `_get_boto3_session`
        method to suit your needs.
        """
        assert bucket is not None, 'S3 Bucket must be specified using the `bucket` kwarg.'
        self.bucket = bucket
        self.aws_profile = aws_profile
        self.use_opinel = use_opinel
        self.skip_hadoop_artifacts = skip_hadoop_artifacts
        self.__path_separator = path_separator
        self._session = session
        self._client = None

        # Ensure self.host is updated with correct AWS region
        import boto3
        self.host = 'autoscaling.{}.amazonaws.com'.format(
            (session or boto3.Session(profile_name=self.aws_profile)).region_name or 'us-east-1'
        )

        # Mask logging from botocore's vendored libraries
        logging.getLogger('botocore.vendored').setLevel(100)

    @override
    def _connect(self):
        self._session = self._session or self._get_boto3_session()
        self._client = self._session.client('s3')
        self._resource = self._session.resource('s3')

    def _get_boto3_session(self):
        import boto3

        if self.use_opinel:
            from opinel.utils.credentials import read_creds

            # Refresh access token, and attach credentials to current object for debugging
            self._credentials = read_creds(self.aws_profile)

            return boto3.Session(
                aws_access_key_id=self._credentials['AccessKeyId'],
                aws_secret_access_key=self._credentials['SecretAccessKey'],
                aws_session_token=self._credentials['SessionToken'],
                profile_name=self.aws_profile,
            )

        return boto3.Session(profile_name=self.aws_profile)

    @override
    def _is_connected(self):
        if self._client is None:
            return False
        # Check if still able to perform requests against AWS
        import botocore
        try:
            self._client.list_buckets()
        except botocore.exceptions.ClientError as e:
            if len(e.args) > 0:
                if 'ExpiredToken' in e.args[0] or 'InvalidToken' in e.args[0]:
                    return False
                elif 'AccessDenied' in e.args[0]:
                    return True

    @override
    def _disconnect(self):
        pass

    # Path properties and helpers
    @override
    def _path_home(self):
        return self.path_separator

    @override
    def _path_separator(self):
        return self.__path_separator

    # File node properties
    @override
    def _exists(self, path):
        return self.isfile(path) or self.isdir(path)

    def _s3_path(self, path):
        if path.startswith(self.path_separator):
            path = path[len(self.path_separator):]
        if path.endswith(self.path_separator):
            path = path[:-len(self.path_separator)]
        return path

    @override
    def _isdir(self, path):
        response = next(iter(self.__dir_paginator(path)))
        if 'CommonPrefixes' in response or 'Contents' in response:
            return True
        return False

    @override
    def _isfile(self, path):
        try:
            self._client.get_object(Bucket=self.bucket, Key=self._s3_path(path) or '')
            return True
        except:
            return False

    # Directory handling and enumeration

    def __dir_paginator(self, path):
        path = self._s3_path(path)
        paginator = self._client.get_paginator('list_objects')
        iterator = paginator.paginate(
            Bucket=self.bucket,
            Prefix=path + (self.path_separator if path else ''),
            Delimiter=self.path_separator,
            PaginationConfig={'PageSize': 500}
        )
        return iterator

    @override
    def _dir(self, path):
        iterator = self.__dir_paginator(path)

        for response_data in iterator:
            for prefix in response_data.get('CommonPrefixes', []):
                yield FileSystemFileDesc(
                    fs=self,
                    path=prefix['Prefix'][:-len(self.path_separator)],
                    name=prefix['Prefix'][:-len(self.path_separator)].split(self.path_separator)[-1],  # Remove trailing slash
                    type='directory',
                )
            for prefix in response_data.get('Contents', []):
                if self.skip_hadoop_artifacts and prefix['Key'].endswith('_$folder$'):
                    continue
                yield FileSystemFileDesc(
                    fs=self,
                    path=prefix['Key'],
                    name=prefix['Key'].split(self.path_separator)[-1],
                    type='file',
                    bytes=prefix['Size'],
                    owner=prefix['Owner']['DisplayName'] if 'Owner' in prefix else None,
                    last_modified=prefix['LastModified']
                )

    # TODO: Interestingly, directly using Amazon S3 methods seems slower than generic approach. Hypothesis: keys is not asynchronous.
    # def _find(self, path_prefix, **attrs):
    #     if len(set(attrs).difference(('name',))) > 0 or hasattr(attrs.get('name'), '__call__'):
    #         logger.warning('Falling back to recursive search, rather than using S3, since find requires filters on more than just name.')
    #         for result in super(S3Client, self)._find(path_prefix, **attrs):
    #             yield result
    #
    #     pattern = re.compile(attrs.get('name') or '.*')
    #
    #     b = self._resource.Bucket(self.bucket)
    #     keys = b.objects.filter(Prefix=path_prefix)
    #     for k in keys:
    #         if pattern is None or pattern.match(k.key[len(path_prefix):]):
    #             yield k.key

    @override
    def _mkdir(self, path, recursive, exist_ok):
        path = self._s3_path(path)
        if not path.endswith(self.path_separator):
            path += self.path_separator
        if not self._exists(path):
            self._client.put_object(Bucket=self.bucket, Key=path)

    @override
    def _remove(self, path, recursive):
        path = self._s3_path(path)
        if recursive:
            bucket = self._resource.Bucket(self.bucket)
            to_delete = []
            for obj in bucket.objects.filter(Prefix=path + self.path_separator):
                to_delete.append({'Key': obj.key})
                if len(to_delete) == 1000:  # Maximum number of simultaneous deletes is 1000
                    self._client.delete_objects(Bucket=self.bucket, Delete={'Objects': to_delete})
                    to_delete = []
            self._client.delete_objects(Bucket=self.bucket, Delete={'Objects': to_delete})
        self._client.delete_object(Bucket=self.bucket, Key=path)

    # File handling
    @override
    def _file_read_(self, path, size=-1, offset=0, binary=False):
        if not self.isfile(path):
            raise FileNotFoundError("File `{}` does not exist.".format(path))

        obj = self._resource.Object(self.bucket, self._s3_path(path))
        body = obj.get()['Body'].read()

        if not binary:
            body = body.decode('utf-8')
        if offset > 0:
            body = body[offset:]
        if size >= 0:
            body = body[:size]
        return body

    @override
    def _file_append_(self, path, s, binary):
        raise NotImplementedError("Support for S3 append operation has yet to be implemented.")

    @override
    def _file_write_(self, path, s, binary):
        obj = self._resource.Object(self.bucket, self._s3_path(path))
        if not binary:
            s = s.encode('utf-8')
        obj.put(Body=s)
        return True



import getpass
import re
from abc import abstractmethod

import six
from interface_meta import quirk_docs, override
from future.utils import raise_with_traceback

from omniduct.duct import Duct
from omniduct.errors import DuctAuthenticationError, DuctServerUnreachable
from omniduct.filesystems.base import FileSystemClient
from omniduct.utils.decorators import require_connection
from omniduct.utils.ports import get_free_local_port, is_local_port_free

try:  # Python 3
    from urllib.parse import urlparse, urlunparse
except ImportError:  # Python 2
    from urlparse import urlparse, urlunparse


class PortForwardingRegister(object):
    """
    A register of all port forwards initiated by a particular Duct.
    """

    def __init__(self):
        self._register = {}

    def lookup(self, remote_host, remote_port):
        """
        Look up a previously forwarded remote port.

        Args:
            remote_host (str): The remote host.
            remote_port (int): The remote port.

        Returns:
            tuple, None: A tuple of local port and implementation-specific
                connection artifact, if it exists, and `None` otherwise.
        """
        return self._register.get('{}:{}'.format(remote_host, remote_port))

    def lookup_port(self, remote_host, remote_port):
        """
        Look up the local port bound to a remote port.

        Args:
            remote_host (str): The remote host.
            remote_port (int): The remote port.

        Returns:
            int, None: The local port use in the port forward, or `None` if
                port forward does not exist.
        """
        entry = self.lookup(remote_host, remote_port)
        if entry is not None:
            return entry[0]

    def reverse_lookup(self, local_port):
        """
        Look up a remote host / port associated with a local port.

        Args:
            local_port (int): The local port.

        Returns:
            list: A list of remote hostname, remote port, and
                implementation-specific connection artifact.
        """
        for key, (port, connection) in self._register.items():
            if port == local_port:
                return key.split(':') + [connection]
        return None

    def register(self, remote_host, remote_port, local_port, connection):
        """
        Register a port-forward connection.

        Args:
            remote_host (str): The remote host.
            remote_port (int): The remote port.
            local_port (int): The local port.
            connection (object): Implementation-specific connection artifact.
        """
        key = '{}:{}'.format(remote_host, remote_port)
        if key in self._register:
            raise RuntimeError("Remote host/port combination ({}) is already registered.".format(key))
        self._register[key] = (local_port, connection)

    def deregister(self, remote_host, remote_port):
        """
        Deregister a port-forward connection.

        Args:
            remote_host (str): The remote host.
            remote_port (int): The remote port.

        Returns:
            tuple: A tuple of local port and implementation-specific
                connection artifact, if it exists, and `None` otherwise.
        """
        return self._register.pop('{}:{}'.format(remote_host, remote_port))


class RemoteClient(FileSystemClient):
    """
    An abstract class providing the common API for all remote clients.

    Attributes:
        smartcard (dict): Mapping of smartcard names to system libraries
            compatible with `ssh-add -s '<system library>' ...`.
    """
    __doc_attrs = """
    smartcard (dict): Mapping of smartcard names to system libraries
        compatible with `ssh-add -s '<system library>' ...`.
    """

    DUCT_TYPE = Duct.Type.REMOTE
    DEFAULT_PORT = None

    @quirk_docs('_init', mro=True)
    def __init__(self, smartcards=None, **kwargs):
        """
        Args:
            smartcards (dict): Mapping of smartcard names to system libraries
                compatible with `ssh-add -s '<system library>' ...`.
        """

        self.smartcards = smartcards
        self.__port_forwarding_register = PortForwardingRegister()

        FileSystemClient.__init_with_kwargs__(self, kwargs, port=self.DEFAULT_PORT)

        # Note: self._init is called by FileSystemClient constructor.

    @override
    @abstractmethod
    def _init(self):
        raise NotImplementedError

    # SSH commands
    @override
    def connect(self):
        """
        Connect to the remote server.

        It is not normally necessary for a user to manually call this function,
        since when a connection is required, it is automatically created.

        Compared to base `Duct.connect`, this method will automatically catch
        the first `DuctAuthenticationError` error triggered by `Duct.connect`,
        and (if smartcards have been configured) attempt to re-initialise the
        smartcards before trying once more.

        Returns:
            `Duct` instance: A reference to the current object.
        """
        try:
            Duct.connect(self)
        except DuctServerUnreachable as e:
            raise_with_traceback(e)
        except DuctAuthenticationError as e:
            if self.smartcards and self.prepare_smartcards():
                Duct.connect(self)
            else:
                raise_with_traceback(e)
        return self

    def prepare_smartcards(self):
        """
        Prepare smartcards for use in authentication.

        This method checks attempts to ensure that the each of the nominated
        smartcards is available and prepared for use. This may result in
        interactive requests for pin confirmation, depending on the card.

        Returns:
            bool: Returns `True` if at least one smartcard was activated, and
                `False` otherwise.
        """

        smartcard_added = False

        for name, filename in self.smartcards.items():
            smartcard_added |= self._prepare_smartcard(name, filename)

        return smartcard_added

    def _prepare_smartcard(self, name, filename):
        import pexpect

        remover = pexpect.spawn('ssh-add -e "{}"'.format(filename))
        i = remover.expect(["Card removed:", "Could not remove card", pexpect.TIMEOUT])
        if i == 2:
            raise RuntimeError("Unable to reset card using ssh-agent. Output of ssh-agent was: \n{}\n\n"
                               "Please report this error!".format(remover.before))

        adder = pexpect.spawn('ssh-add -s "{}" -t 14400'.format(filename))
        i = adder.expect(['Enter passphrase for PKCS#11:', pexpect.TIMEOUT])
        if i == 0:
            adder.sendline(getpass.getpass('Please enter your passcode to unlock your "{}" smartcard: '.format(name)))
        else:
            raise RuntimeError("Unable to add card using ssh-agent. Output of ssh-agent was: \n{}\n\n"
                               "Please report this error!".format(remover.before))
        i = adder.expect(['Card added:', pexpect.TIMEOUT])
        if i != 0:
            raise RuntimeError("Unexpected error while adding card. Check your passcode and try again.")

        return True

    @quirk_docs('_execute')
    @require_connection
    def execute(self, cmd, **kwargs):
        """
        Execute a command on the remote server.

        Args:
            cmd (str): The command to run on the remote associated with this
                instance.
            **kwargs (dict): Additional keyword arguments to be passed on to
                `._execute`.

        Returns:
            SubprocessResults: The result of the execution.
        """
        return self._execute(cmd, **kwargs)

    @abstractmethod
    def _execute(self, cmd, **kwargs):
        raise NotImplementedError

    # Port forwarding code

    def _extract_host_and_ports(self, remote_host, remote_port, local_port):
        assert remote_host is None or isinstance(remote_host, six.string_types), "Remote host, if specified, must be a string of form 'hostname(:port)'."
        assert remote_port is None or isinstance(remote_port, int), "Remote port, if specified, must be an integer."
        assert local_port is None or isinstance(local_port, int), "Local port, if specified, must be an integer."

        host = port = None
        if remote_host is not None:
            m = re.match(r'(?P<host>[a-zA-Z0-9\-.]+)(?::(?P<port>[0-9]+))?', remote_host)
            assert m, "Host not valid: {}. Must be a string of form 'hostname(:port)'.".format(remote_host)

            host = m.group('host')
            port = m.group('port') or remote_port
        return host, port, local_port

    @quirk_docs('_port_forward_start')
    @require_connection
    def port_forward(self, remote_host, remote_port=None, local_port=None):
        """
        Initiate a port forward connection.

        This method establishes a local port forwarding from a local port `local`
        to remote port `remote`. If `local` is `None`, an available local port is
        automatically chosen. If the remote port is already forwarded, a new
        connection is not established.

        Args:
            remote_host (str): The hostname of the remote host in form:
                'hostname(:port)'.
            remote_port (int, None): The remote port of the service.
            local_port (int, None): The port to use locally (automatically
                determined if not specified).

        Returns:
            int: The local port which is port forwarded to the remote service.
        """

        # Hostname and port extraction
        remote_host, remote_port, local_port = self._extract_host_and_ports(remote_host, remote_port, local_port)
        assert remote_host is not None, "Remote host must be specified."
        assert remote_port is not None, "Remote port must be specified."

        # Actual port forwarding
        registered_port = self.__port_forwarding_register.lookup_port(remote_host, remote_port)
        if registered_port is not None:
            if local_port is not None and registered_port != local_port:
                self.port_forward_stop(registered_port)
            else:
                return registered_port

        if local_port is None:
            local_port = get_free_local_port()
        else:
            assert is_local_port_free(local_port), "Specified local port is in use, and cannot be used."

        if not self.is_port_bound(remote_host, remote_port):
            raise DuctServerUnreachable("Server specified for port forwarding ({}:{}) is unreachable via '{}' ({}).".format(remote_host, remote_port, self.name, self.__class__.__name__))
        connection = self._port_forward_start(local_port, remote_host, remote_port)
        self.__port_forwarding_register.register(remote_host, remote_port, local_port, connection)

        return local_port

    def has_port_forward(self, remote_host=None, remote_port=None, local_port=None):
        """
        Check whether a port forward connection exists.

        Args:
            remote_host (str): The hostname of the remote host in form:
                'hostname(:port)'.
            remote_port (int, None): The remote port of the service.
            local_port (int, None): The port used locally.

        Returns:
            bool: Whether a port-forward for this remote service exists, or if
                local port is specified, whether that port is locally used for
                port forwarding.
        """
        # Hostname and port extraction
        remote_host, remote_port, local_port = self._extract_host_and_ports(remote_host, remote_port, local_port)

        assert remote_host is not None and remote_port is not None or local_port is not None, "Either remote host and port must be specified, or the local port must be specified."

        if remote_host is not None and remote_port is not None:
            return self.__port_forwarding_register.lookup(remote_host, remote_port) is not None
        else:
            return self.__port_forwarding_register.reverse_lookup(local_port) is not None

    @quirk_docs('_port_forward_stop')
    def port_forward_stop(self, local_port=None, remote_host=None, remote_port=None):
        """
        Disconnect an existing port forward connection.

        If a local port is provided, then the forwarding (if any) associated
        with that port is found and stopped; otherwise any established port
        forwarding associated with the nominated remote service is stopped.

        Args:
            remote_host (str): The hostname of the remote host in form:
                'hostname(:port)'.
            remote_port (int, None): The remote port of the service.
            local_port (int, None): The port used locally.
        """
        # Hostname and port extraction
        remote_host, remote_port, local_port = self._extract_host_and_ports(remote_host, remote_port, local_port)

        assert remote_host is not None and remote_port is not None or local_port is not None, "Either remote host and port must be specified, or the local port must be specified."

        if remote_host is not None and remote_port is not None:
            local_port, connection = self.__port_forwarding_register.lookup(remote_host, remote_port)
        else:
            remote_host, remote_port, connection = self.__port_forwarding_register.reverse_lookup(local_port)

        self._port_forward_stop(local_port, remote_host, remote_port, connection)
        self.__port_forwarding_register.deregister(remote_host, remote_port)

    def port_forward_stopall(self):
        """
        Disconnect all existing port forwarding connections.
        """
        for remote_host in self.__port_forwarding_register._register.copy():
            self.port_forward_stop(remote_host=remote_host)

    def get_local_uri(self, uri):
        """
        Convert a remote uri to a local one.

        This method takes a remote service uri accessible to the remote host and
        returns a local uri accessible directly on the local host, establishing
        any necessary port forwarding in the process.

        Args:
            uri (str): The remote uri to be made local.

        Returns:
            str: A local uri that tunnels all traffic to the remote host.
        """
        parsed_uri = urlparse(uri)
        return urlunparse(parsed_uri._replace(netloc='localhost:{}'.format(self.port_forward(parsed_uri.netloc))))

    def show_port_forwards(self):
        """
        Print to stdout the active port forwards associated with this client.
        """
        if len(self.__port_forwarding_register._register) == 0:
            print("No port forwards currently in use.")
        for remote_host, (local_port, _) in self.__port_forwarding_register._register.items():
            print("localhost:{}".format(local_port), "->", remote_host, "(on {})".format(self._host))

    @abstractmethod
    def _port_forward_start(self, local_port, remote_host, remote_port):
        raise NotImplementedError

    @abstractmethod
    def _port_forward_stop(self, local_port, remote_host, remote_port, connection):
        raise NotImplementedError

    @quirk_docs('_is_port_bound')
    @require_connection
    def is_port_bound(self, host, port):
        """
        Check whether a port on a remote host is accessible.

        This method checks to see whether a particular port is active on a
        given host by attempting to establish a connection with it.

        Args:
            host (str): The hostname of the target service.
            port (int): The port of the target service.

        Returns:
            bool: Whether the port is active and accepting connections.
        """
        return self._is_port_bound(host, port)

    @abstractmethod
    def _is_port_bound(self, host, port):
        pass



import pytest

from omniduct.databases._namespaces import ParsedNamespaces


class TestParseNamespaces:

    def test_simple(self):
        namespace = ParsedNamespaces.from_name(
            name='my_db.my_table',
            namespaces=['database', 'table']
        )

        assert namespace.database == 'my_db'
        assert namespace.table == 'my_table'
        assert namespace.as_dict() == {
            'database': 'my_db',
            'table': 'my_table'
        }

    def test_quoted_names(self):
        namespace = ParsedNamespaces.from_name(
            name='`my_db`.`my . table`',
            namespaces=['catalog', 'database', 'table'],
            quote_char='`'
        )

        assert namespace.catalog is None
        assert namespace.database == 'my_db'
        assert namespace.table == 'my . table'
        assert namespace.as_dict() == {
            'catalog': None,
            'database': 'my_db',
            'table': 'my . table'
        }

    def test_separator(self):
        namespace = ParsedNamespaces.from_name(
            name='cat|my_db|my_table',
            namespaces=['catalog', 'database', 'table'],
            separator='|'
        )

        assert namespace.catalog == 'cat'
        assert namespace.database == 'my_db'
        assert namespace.table == 'my_table'
        assert namespace.as_dict() == {
            'catalog': 'cat',
            'database': 'my_db',
            'table': 'my_table'
        }

    def test_parsing_failure(self):
        with pytest.raises(ValueError):
            ParsedNamespaces.from_name(
                name='my_db.my_table',
                namespaces=['table']
            )

    def test_nonexistent_namespace(self):
        with pytest.raises(AttributeError):
            ParsedNamespaces.from_name(
                name='my_table',
                namespaces=['table']
            ).database

    def test_not_encapsulated(self):
        namespace = ParsedNamespaces.from_name('my_db.my_table', ['database', 'table'])
        assert namespace.as_dict() == {'database': 'my_db', 'table': 'my_table'}

        with pytest.raises(ValueError):
            ParsedNamespaces.from_name(namespace, ['schema', 'table'])

    def test_empty(self):
        namespace = ParsedNamespaces.from_name("", ['database', 'table'])

        assert bool(namespace) is False
        assert namespace.database is None
        assert namespace.table is None
        assert str(namespace) == ''
        assert repr(namespace) == 'Namespace<>'

    def test_parent(self):
        namespace = ParsedNamespaces.from_name(
            name='my_db.my_table',
            namespaces=['catalog', 'database', 'table']
        )

        assert namespace.parent.name == '"my_db"'
        assert namespace.parent.as_dict() == {
            'catalog': None,
            'database': 'my_db'
        }

    def test_casting(self):
        namespace = ParsedNamespaces.from_name(
            name='my_db.my_table',
            namespaces=['catalog', 'database', 'table']
        )

        assert str(namespace) == '"my_db"."my_table"'
        assert bool(namespace) is True
        assert namespace.__bool__() == namespace.__nonzero__()  # Python 2/3 compatibility
        assert repr(namespace) == 'Namespace<"my_db"."my_table">'
